import numpy as np
from extendedstim.Code.QuantumCode.PauliCode import PauliCode
from extendedstim.Code.QuantumCode.QuantumCSSCode import QuantumCSSCode
from extendedstim.Math.BinaryArray import BinaryArray as ba
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator


class PauliCSSCode(PauliCode, QuantumCSSCode):

    # %%  USER：构造方法
    def __init__(self, generators_x, generators_z, physical_number):
        QuantumCSSCode.__init__(self, generators_x, generators_z, physical_number)

    # %%  USER：属性方法
    ##  TODO：求Pauli CSS code的距离（x方向）
    @property
    def distance_x(self):
        return 1

    ##  TODO：求Pauli CSS code的距离（z方向）
    @property
    def distance_z(self):
        return 1

    ##  TODO：求Pauli CSS code的逻辑算子（x方向）
    @property
    def logical_operators_x(self):
        if self._logical_operators_x is not None:
            return self._logical_operators_x
        else:
            raise NotImplementedError('这个功能还没有实现')

    ##  TODO：求Pauli CSS code的逻辑算子（z方向）
    @property
    def logical_operators_z(self):
        if self._logical_operators_z is not None:
            return self._logical_operators_z
        else:
            raise NotImplementedError('这个功能还没有实现')

    @staticmethod
    def SteaneCode():
        generators_x = [PauliOperator([3,4,5,6],[],1),PauliOperator([1,2,5,6],[],1),PauliOperator([0,2,4,6],[],1)]
        generators_z = [PauliOperator([],[3,4,5,6],1),PauliOperator([],[1,2,5,6],1),PauliOperator([],[0,2,4,6],1)]
        physical_number=7
        result=PauliCSSCode(generators_x, generators_z, physical_number)
        result._logical_operators_x=[PauliOperator([0,1,2],[],1)]
        result._logical_operators_z=[PauliOperator([],[0,1,2],1)]

    @staticmethod
    def SurfaceCode(d):
        generators_x=[]
        generators_z=[]
        if d==3:
            generators_x=[
                PauliOperator([6,7],[],1),
                PauliOperator([4,5,7,8],[],1),
                PauliOperator([0,1,3,4],[],1),
                PauliOperator([1,2],[],1)
            ]
            generators_z=[
                PauliOperator([],[3,4,6,7],1),
                PauliOperator([],[5,8],1),
                PauliOperator([],[0,3],1),
                PauliOperator([],[1,2,4,5],1)
            ]
            physical_number=9
            result=PauliCSSCode(generators_x, generators_z, physical_number)
            result._logical_operators_x=[PauliOperator([0,3,6],[],1)]
            result._logical_operators_z=[PauliOperator([],[0,1,2],1)]
        return result



import copy
from abc import ABC, abstractmethod
import numpy as np
from extendedstim.Physics.Operator import Operator


class QuantumCode(ABC):

    #%%  USER：构造方法
    def __init__(self,generators,physical_number):
        assert isinstance(generators, list) or isinstance(generators, np.ndarray)
        self.generators = np.array(generators,dtype=Operator)
        self.physical_number=physical_number
        self.checker_number=self.generators.shape[0]

    #%%  USER：属性方法
    ##  USER：求校验矩阵
    @property
    def check_matrix(self):
        matrix=Operator.get_matrix(self.generators,self.physical_number)
        return matrix

    ##  USER：求校验矩阵的秩
    @property
    def rank(self):
        return self.check_matrix.rank

    ##  USER：求logical number
    @property
    def logical_number(self):
        return self.physical_number-self.rank

    ##  USER：求码距
    @property
    @abstractmethod
    def distance(self):
        pass

    ##  USER：求逻辑算符
    @property
    @abstractmethod
    def logical_operators(self):
        pass

    #%%  USER：对象方法
    ##  USER：修改索引
    def index_map(self, index_map):
        for generator in self.generators:
            generator.index_map(index_map)

    ##  USER：复制代码
    def copy(self):
        return copy.deepcopy(self)


import copy
import galois
import numpy as np
from mip import Model, xsum, minimize, BINARY


class BinaryArray:
    GF = galois.GF(2)

    #%%  USER：===构造方法===
    def __init__(self, array):
        if isinstance(array, BinaryArray):
            self._array = copy.deepcopy(array._array)  # 拷贝生成二进制数组
        else:
            self._array = self.GF(np.array(array, dtype=int))  # 转换为二进制数组
        self.shape = self._array.shape  # 数组形状

    #%%  USER：===重载运算符===
    ##  USER：---返回数组的长度---
    def __len__(self):
        return self.shape[0]

    ##  USER：---判断两个数组是否相等---
    def __eq__(self, other):
        if isinstance(other, BinaryArray):
            return self._array == other._array
        else:
            return self._array == other

    ##  USER：---获取数组的元素---
    def __getitem__(self, item):
        result=self._array[item]
        if isinstance(result,BinaryArray):
            return result
        else:
            return BinaryArray(result)

    ##  USER：---设置数组的元素---
    def __setitem__(self, item, value):
        self._array[item] = value

    ##  USER：---返回数组的字符串表示---
    def __str__(self):
        return str(self._array)

    ##  USER：---数组相加---
    def __add__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(self._array + other._array)

    ##  USER：---数组相减---
    def __sub__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(self._array - other._array)

    ##  USER：---数组相乘---
    def __mul__(self, other):
        assert (isinstance(other, BinaryArray)) or (isinstance(other, int))
        if isinstance(other, int):
            return BinaryArray(self._array * other)
        else:
            return BinaryArray(self._array * other._array)

    ##  USER：---数组右乘---
    def __rmul__(self, other):
        assert (isinstance(other, BinaryArray)) or (isinstance(other, int))
        if isinstance(other, int):
            return BinaryArray(self._array * other)
        else:
            return BinaryArray(self._array * other._array)

    ##  USER：---数组矩阵乘---
    def __matmul__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(self._array @ other._array)

    ##  USER：---数组右矩阵乘---
    def __rmatmul__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(other._array @ self._array)

    ##  USER：---数组幂运算---
    def __pow__(self, power):
        return BinaryArray(np.linalg.matrix_power(self._array, power))

    #%%  USER：===属性方法===
    ##  USER：---返回数组的转置---
    @property
    def T(self):
        return BinaryArray(self._array.T)

    ##  USER：---求占据的位置---
    @property
    def occupy(self):
        return np.where(self._array != 0)[0]

    ##  USER：---返回数组的权重---
    @property
    def weight(self):
        return np.count_nonzero(self._array)

    ##  USER：---返回数组的零空间---
    @property
    def null_space(self):
        return BinaryArray(self._array.null_space())

    ##  USER：---返回数组的秩---
    @property
    def rank(self):
        return np.linalg.matrix_rank(self._array)

    #%%  USER：===对象方法===
    ##  USER：---复制函数---
    def copy(self):
        return copy.deepcopy(self)

    #%%  USER：===静态方法===
    ##  USER：---矩阵求和---
    @staticmethod
    def sum(array_list):
        assert (isinstance(array_list, list))
        assert (isinstance(array_list[0], BinaryArray))
        return BinaryArray(np.sum([array._array for array in array_list], axis=0))

    ##  USER：---求数组的水平堆叠---
    @staticmethod
    def hstack(array, *args):
        assert len(args) > 0
        assert isinstance(array, BinaryArray)
        for temp in args:
            assert isinstance(temp, BinaryArray)
        if len(array)!=0:
            result = array._array
        else:
            result=[]
        for temp in args:
            if len(temp)!=0:
                if len(result)==0:
                    result = temp._array
                else:
                    result = np.hstack((result, temp._array))
        return BinaryArray(result)

    ##  USER：---求数组的垂直堆叠---
    @staticmethod
    def vstack(array, *args):
        assert len(args) > 0
        assert isinstance(array, BinaryArray)
        for temp in args:
            assert isinstance(temp, BinaryArray)
        if len(array)!=0:
            result = array._array
        else:
            result=[]
        for temp in args:
            if len(temp)!=0:
                if len(result)==0:
                    result = temp._array
                else:
                    result = np.vstack((result, temp._array))
        return BinaryArray(result)

    ##  USER：---求数组的点积---
    @staticmethod
    def dot(a,b):
        assert (isinstance(a, BinaryArray))
        assert (isinstance(b, BinaryArray))
        return np.dot(a._array,b._array)

    ##  USER：---求解线性系统---
    @staticmethod
    def solve(matrix, vector):
        assert isinstance(matrix, BinaryArray)
        assert isinstance(vector, BinaryArray)
        assert matrix.shape[1] == vector.shape[0]

        ##  使用GF(2)数组
        A = matrix._array.T
        b = vector._array.reshape(-1, 1)

        # 构造增广矩阵
        aug = np.concatenate((A, b), axis=1)
        n, m_plus_1 = aug.shape
        m = m_plus_1 - 1
        rank = 0

        ##  前向消元
        for col in range(m):
            ##  寻找主元
            pivot_row = None
            for i in range(rank, n):
                if aug[i, col] == 1:
                    pivot_row = i
                    break

            if pivot_row is None:
                ##  该列没有主元，跳过
                continue

            ##  交换行
            if pivot_row != rank:
                aug[[rank, pivot_row], :] = aug[[pivot_row, rank], :]

            ##  消去当前列下方和上方的元素
            for i in range(n):
                if i != rank and aug[i, col] == 1:
                    aug[i, :] ^= aug[rank, :]
            rank += 1

        ##  检查是否有解
        for i in range(rank, n):
            if aug[i, -1] == 1:
                ##  矛盾方程，无解
                return None

        ##  构造解向量
        solution = BinaryArray.GF.Zeros(m)
        leading_cols = []

        ##  找出主元列
        for i in range(rank):
            for j in range(m):
                if aug[i, j] == 1:
                    leading_cols.append(j)
                    break

        ##  回代求解
        for i in range(rank):
            col = leading_cols[i]
            solution[col] = aug[i, -1]

            ##  消去当前行中主元列右侧的元素对解的影响
            for j in range(col + 1, m):
                if aug[i, j] == 1:
                    solution[col] ^= solution[j]

        ##  返回解向量
        return BinaryArray(solution)

    ##  USER：---返回两个线性空间的交集---
    @staticmethod
    def cap(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)
        GF2 = galois.GF(2)

        ##  转化为GF2数组
        matrix1 = matrix1._array
        matrix2 = matrix2._array
        m = matrix1.shape[0]
        k = matrix2.shape[0]

        ##  解方程 basis1^T * x + basis2^T * y = 0 (表示交集向量), 构造矩阵 [basis1^T | basis2^T]
        aug_matrix = GF2(np.concatenate((matrix1.T, matrix2.T), axis=1))

        ##  计算零空间（解空间）
        nullspace = aug_matrix.null_space()

        ##  从零空间中取对应于每个解向量的系数（对应于方程组中的解向量）
        ab_space = nullspace[:, :m]

        ##  将系数乘以 basis1 得到具体解（交集）
        if len(ab_space) == 0:
            return GF2.Zeros((0, matrix1.shape[1]))

        intersection_vectors = ab_space @ matrix1

        ##  产生基向量的线性无关组合，即每种向量的不同的高斯行组合样本
        rref_intersection = intersection_vectors.row_reduce()

        ##  除去全0行的行（即为没有有用基底捕获的时候）
        nz_mask = np.any(rref_intersection != 0, axis=1)
        rref_basis = rref_intersection[nz_mask]

        ##  返回补集
        return BinaryArray(rref_basis)

    ##  USER：---返回两个线性空间的差集---
    @staticmethod
    def minus(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)

        ##  转化为GF2数组
        matrix1_origin = matrix1
        matrix2_origin = matrix2
        matrix1 = matrix1._array
        matrix2 = matrix2._array

        intersect = BinaryArray.cap(matrix1_origin, matrix2_origin)._array
        if len(intersect) == 0:
            return matrix1

        result = []
        for i in range(len(matrix1)):
            rank = np.linalg.matrix_rank(intersect)
            intersect = np.vstack((intersect, matrix1[i]))
            if np.linalg.matrix_rank(intersect) > rank:
                result.append(matrix1[i])

        ##  返回差集
        return BinaryArray(result)

    ##  USER：---返回两个线性空间的直积---
    @staticmethod
    def direct_sum(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)

        ##  转化为GF2数组
        matrix1 = matrix1._array
        matrix2 = matrix2._array
        result = matrix1[0].copy()

        ##  直接拼接
        for i in range(1, len(matrix1)):
            rank = np.linalg.matrix_rank(result)
            temp = np.vstack((result, matrix1[i]))
            if np.linalg.matrix_rank(temp) > rank:
                result = temp
        for i in range(len(matrix2)):
            rank = np.linalg.matrix_rank(result)
            temp = np.vstack((result, matrix2[i]))
            if np.linalg.matrix_rank(temp) > rank:
                result = temp

        ##  返回结果
        return BinaryArray(result)

    ##  USER：---正交化基矢组---
    @staticmethod
    def orthogonalize(matrix):
        def gf2_gram_schmidt(vectors, bilinear_form):
            k = len(vectors)
            ##  复制向量以避免修改原始数据
            B_i = [v.copy() for v in vectors]
            O = []  # 存储正交基
            for i in range(k):
                if np.mod(np.count_nonzero(B_i[i]), 2) == 1:
                    for j in range(k):
                        if j != i and np.mod(np.count_nonzero(B_i[j]), 2) == 0:
                            B_i[j] = B_i[j] + B_i[i]
                    break
                if i == k - 1:
                    print('There is no odd-weight vector')
                    return False

            while True:
                length = 10000
                b1 = None
                flag = 0
                b1 = B_i[0]
                # for i in range(len(B_i)):
                #     if length > np.count_nonzero(B_i[i]):
                #         b1 = B_i[i]
                #         flag = i
                #         length = np.count_nonzero(B_i[i])

                o_i = b1

                O.append(o_i)

                next_B = []
                for j in range(len(B_i)):
                    if j != flag:
                        b = B_i[j]
                        coef = bilinear_form(b, o_i)
                        b_new = b + coef * o_i
                        next_B.append(b_new)

                B_i = next_B
                if len(B_i) == 0:
                    return O
                length = len(B_i)
                for i in range(length):
                    if np.mod(np.count_nonzero(B_i[i]), 2) == 1:
                        for j in range(length):
                            if j != i and np.mod(np.count_nonzero(B_i[j]), 2) == 0:
                                B_i[j] = B_i[j] + B_i[i]
                        break
                    if i == len(B_i):
                        return O

        ##  双线性形式矩阵
        bilinear_form1 = lambda u, v: np.dot(u, v)
        matrix_GF=[temp._array for temp in matrix]
        ortho_basis1 = gf2_gram_schmidt(matrix_GF, bilinear_form1)

        ##  输出结果
        return BinaryArray(ortho_basis1)

    ##  USER：---计算张量积---
    @staticmethod
    def kron(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)
        return BinaryArray(np.kron(matrix1._array, matrix2._array))

    ##  USER：---计算code distance---
    @staticmethod
    def distance(H, method):
        assert isinstance(H, BinaryArray)
        if method == 'mip':
            logic_op = BinaryArray.minus(H.null_space, H)
            return mip_distance_caculator(H._array, logic_op._array)
        elif method == 'random':
            return random_distance_caculator(H._array, H._array, 500, mindist=1, debug=1, field=2)

    ##  USER：---计算子系统距离---
    @staticmethod
    def subsystem_distance(stabilizers, gauges, method):
        pass

    ##  USER：---根据占用位置生成二进制数组---
    @staticmethod
    def FromOccupy(occupy, *args):
        if len(args) == 0:
            assert isinstance(occupy, list) or isinstance(occupy, np.ndarray)
            temp = np.zeros(np.max(occupy) + 1)
            temp[occupy] = 1
            return BinaryArray(temp)
        elif len(args) == 1:
            assert isinstance(occupy, list) or isinstance(occupy, np.ndarray)
            assert isinstance(args[0], int)
            temp = np.zeros(args[0], dtype=int)
            temp[occupy] = 1
            return BinaryArray(temp)
        else:
            raise ValueError("FromOccupy() takes 1 or 2 positional arguments but {} were given".format(len(args) + 1))

    ##  USER：---根据数组生成二进制数组---
    @staticmethod
    def FromArray(array):
        return BinaryArray(array)

    ##  USER：---生成全零数组---
    @staticmethod
    def zeros(number):
        return BinaryArray(np.zeros(number, dtype=int))

    ##  USER：---生成单位矩阵---
    @staticmethod
    def eye(number):
        return BinaryArray(np.eye(number, dtype=int))

    ##  USER：---生成循环左移矩阵---
    @staticmethod
    def shift(number, shift):
        S = np.zeros((number, number), dtype=int)
        for i in range(number):
            S[i, (i + shift) % number] = 1
        return BinaryArray(S)

    ##  USER：---生成全一数组---
    @staticmethod
    def ones(number):
        return BinaryArray(np.ones(number, dtype=int))


#%%  KEY：===计算code distance===
def mip_distance_caculator(H, logicOp):

    ##  格式化输入
    H = np.array(H, dtype=int)  # 转换为整数类型的numpy数组
    logicOp = np.array(logicOp, dtype=int)  # 转换为整数类型的numpy数组
    d = H.shape[1]  # 初始化距离为量子比特数量（最大可能距离）

    ##  遍历每个逻辑算子
    for i in range(logicOp.shape[0]):
        logicOp_i = logicOp[i, :]
        n = H.shape[1]  # 量子比特数量（稳定子矩阵的列数）
        m = H.shape[0]  # 稳定子数量（稳定子矩阵的行数）
        wstab = np.max([np.sum(H[i, :]) for i in range(m)])  # 计算最大稳定子权重（单个稳定子中非零元素的最大数量）
        wlog = np.count_nonzero(logicOp_i)  # 计算逻辑算子的权重
        num_anc_stab = int(np.ceil(np.log2(wstab)))  # 计算稳定子约束所需的辅助变量数量（基于最大稳定子权重的对数）
        num_anc_logical = int(np.ceil(np.log2(wlog)))  # 计算逻辑算子约束所需的辅助变量数量（基于逻辑算子权重的对数）
        num_var = n + m * num_anc_stab + num_anc_logical  # 总变量数量 = 量子比特变量 + 稳定子辅助变量 + 逻辑算子辅助变量

        ##  创建混合整数规划模型
        model = Model()
        model.verbose = 0  # 关闭详细输出
        x = [model.add_var(var_type=BINARY) for i in range(num_var)]  # 创建二进制变量数组
        model.objective = minimize(xsum(x[i] for i in range(n)))  # 目标函数：最小化前n个变量（量子比特变量）的和（即最小化Hamming权重）

        # 为每个稳定子添加正交性约束（模2）
        for row in range(m):
            weight = [0] * num_var  # 初始化权重向量
            supp = np.nonzero(H[row, :])[0]  # 获取当前稳定子的支持集（非零元素的位置）

            ##  设置qubit变量的权重为1
            for q in supp:
                weight[q] = 1

            ##  添加辅助变量来处理模2约束
            cnt = 1
            for q in range(num_anc_stab):
                weight[n + row * num_anc_stab + q] = -(1 << cnt)  # 设置辅助变量的权重为负的2的幂次方
                cnt += 1
            model += xsum(weight[i] * x[i] for i in range(num_var)) == 0  # 添加约束：权重向量与变量向量的点积等于0

        ##  添加逻辑算子的奇数重叠约束
        supp = np.nonzero(logicOp_i)[0]  # 获取逻辑算子的支持集
        weight = [0] * num_var  # 初始化权重向量

        ##  设置qubit变量的权重为1
        for q in supp:
            weight[q] = 1

        ##  添加辅助变量来处理模2约束
        cnt = 1
        for q in range(num_anc_logical):
            # 设置辅助变量的权重为负的2的幂次方
            weight[n + m * num_anc_stab + q] = -(1 << cnt)
            cnt += 1

        ##  添加约束：权重向量与变量向量的点积等于1（奇数重叠）
        model += xsum(weight[i] * x[i] for i in range(num_var)) == 1

        ##  求解优化问题，计算最优解中前n个变量的和，即最小Hamming权重
        model.optimize()
        # noinspection PyTypeChecker
        opt_val = sum([x[i].x for i in range(n)])
        d = min(d, int(opt_val))

    ##  返回最小距离
    return d


#%%  KEY：===计算随机距离===
def random_distance_caculator(gx, gz, num, mindist=2, debug=0, field=None):
    ##  设置默认有限域为GF(2)（二进制域）
    if field is None:
        F = galois.GF(2)
    else:
        F = galois.GF(field)

    ##  调试模式检查：验证X和Z稳定子是否正交（CSS码的基本要求）
    if debug & 2:
        if np.any(gx @ gz.T):
            raise ValueError("Input matrices must have orthogonal rows")

    wz = F(gx.null_space())  # 计算X稳定子的零空间（Z逻辑算子空间）
    wx = F(gz.null_space())  # 计算Z稳定子的零空间（X逻辑算子空间）
    rows_wz, cols_wz = wz.shape  # 获取零空间矩阵的维度信息
    dist_bound = cols_wz + 1  # 初始化距离上界为最大可能值（列数+1）
    vec_count = 0  # 计数器：记录找到当前最小权重的向量数量

    ##  主循环：进行num次随机迭代
    for i in range(num):
        per = np.random.permutation(cols_wz)  # 生成随机排列，用于随机化搜索顺序
        wz1 = wz[:, per]  # 对零空间矩阵的列进行随机排列
        wz2 = wz1.row_reduce()  # 对排列后的矩阵进行行约简（高斯消元）
        wz2 = wz2[:, np.argsort(per)]  # 将列顺序恢复为原始顺序

        ##  遍历行约简后的每一行
        for j in range(rows_wz):
            temp_vec = wz2[j, :]  # 获取当前行向量
            temp_weight = np.count_nonzero(temp_vec)  # 计算向量的Hamming权重（非零元素个数）

            ##  检查权重是否在有效范围内且小于等于当前最小距离
            if 0 < temp_weight <= dist_bound:

                ##  检查向量是否与X逻辑算子空间有非零重叠（即是否为非平凡逻辑算子）
                if np.any(wx @ temp_vec):

                    ##  如果找到更小的权重，更新最小距离
                    if temp_weight < dist_bound:
                        dist_bound = temp_weight
                        vec_count = 1

                    ##  如果权重等于当前最小距离，增加计数器
                    elif temp_weight == dist_bound:
                        vec_count += 1

            ##  检查是否达到最小距离阈值，如果达到则提前终止
            if dist_bound <= mindist:
                return mindist

    ##  返回找到的最小距离
    return dist_bound


import copy
import numpy as np
from extendedstim.Code.QuantumCode.MajoranaCode import MajoranaCode
from extendedstim.Circuit.Circuit import Circuit
from extendedstim.Code.QuantumCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.QuantumCode.PauliCSSCode import PauliCSSCode
from extendedstim.Code.QuantumCode.PauliCode import PauliCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator
from extendedstim.tools import *


#%%  USER：===将量子码转换为量子线路===
def Code2Circuit(code,p_noise,p_measure,noise_model,cycle_number):
    ##  ---数据预处理---
    assert isreal(p_noise) and 0<=p_noise<=1
    assert isreal(p_measure) and 0<=p_measure<=1
    assert isinteger(cycle_number) and cycle_number>=0
    assert isinstance(code,MajoranaCode) or isinstance(code,PauliCode)

    ##  ---根据量子码类型选择不同的处理函数---
    ##  处理现象级噪声
    if noise_model=='phenomenological':
        if isinstance(code,MajoranaCSSCode):
            return MajoranaCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,PauliCSSCode):
            return PauliCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,MajoranaCode):
            raise NotImplementedError
        elif isinstance(code,PauliOperator):
            raise NotImplementedError
        else:
            raise NotImplementedError

    ##  处理电路级噪声
    elif noise_model=='circuit-level':
        if isinstance(code,MajoranaCSSCode):
            return MajoranaCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,PauliCSSCode):
            return PauliCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,MajoranaCode):
            raise NotImplementedError
        elif isinstance(code,PauliOperator):
            raise NotImplementedError
        else:
            raise NotImplementedError

    ##  处理代码容量级噪声
    elif noise_model=='code-capacity':
        if isinstance(code,MajoranaCSSCode):
            return MajoranaCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,PauliCSSCode):
            return PauliCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number)
        elif isinstance(code,MajoranaCode):
            raise NotImplementedError
        elif isinstance(code,PauliOperator):
            raise NotImplementedError
        else:
            raise NotImplementedError

    ##  其他类型抛出异常
    else:
        raise ValueError('noise_model must be phenomenological, circuit-level, or code-capacity')


#%%  KEY：===将Majorana CSS code转换为现象级噪声下的测试线路===
def MajoranaCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number:int):

    ##  ---数据预处理---
    assert isinstance(code,MajoranaCSSCode)
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z

    ##  获取logical operators
    logical_occupy=[]
    for i in range(len(logical_x)):
        logical_occupy.append(1j*logical_x[i]@logical_z[i])

    majorana_number=code.physical_number
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)

    ##  ---生成线路---
    ##  线路初始化
    circuit_z = Circuit()
    circuit_z.append({"name":"FORCE","majorana_state":np.append(logical_occupy,np.append(stabilizers_x, stabilizers_z)),
                      'pauli_state':[PauliOperator([],[],1)]*(stabilizer_number+len(logical_occupy)),
                      'pauli_number':0,'majorana_number':majorana_number})

    ##  第一轮测量假设完美的初始化
    observable_include= []  # 记录可观测量的索引
    for i,logical_operator in enumerate(logical_occupy):
        circuit_z.append({"name":"MPP","target":logical_operator,"index":i})
        observable_include.append(len(circuit_z.measurements)-1)
    for i,stabilizer in enumerate(stabilizers_x):
        circuit_z.append({"name":"MPP","target":stabilizer,"index":i+len(logical_occupy)})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit_z.append({"name":"MPP","target":stabilizer,"index":i+len(logical_occupy)+len(stabilizers_x)})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  量子噪声信道
        for i in range(majorana_number):
            circuit_z.append({"name":"FDEPOLARIZE1","target":i,"index":i,"p":p_noise})

        ##  测量稳定子
        for i,stabilizer in enumerate(stabilizers_x):
            circuit_z.append({"name":"MPP","target":stabilizer,"index":i+len(logical_occupy),'p':p_measure})
        for i,stabilizer in enumerate(stabilizers_z):
            circuit_z.append({"name":"MPP","target":stabilizer,"index":i+len(logical_occupy)+len(stabilizers_x),'p':p_measure})

        ##  添加探测器
        for i in range(stabilizer_number):
            circuit_z.append({"name":"DETECTOR","target":[-i - 1, -i - stabilizer_number-1]})

    ##  最后一轮测量稳定子，假设没有噪声
    for i,stabilizer in enumerate(stabilizers_x):
        circuit_z.append({"name":"MPP","target":stabilizer,"index":i+len(logical_occupy)})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit_z.append({"name":"MPP","target":stabilizer,"index":i+len(logical_occupy)+len(stabilizers_x)})

    ##  测量逻辑算符
    for i,logical_operator in enumerate(logical_occupy):
        circuit_z.append({"name":"MPP","target":logical_operator,"index":i})
        circuit_z.append({"name":"OBSERVABLE_INCLUDE","target":[len(circuit_z.measurements) - 1, observable_include[i]]})

    ##  ---返回线路---
    return circuit_z


#%%  KEY：===将Majorana CSS code转换为线路级噪声下的测试线路===
def MajoranaCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number:int):

    ##  ---数据预处理---
    ##  获取稳定子
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z

    ##  获取逻辑算符
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z
    logical_occupy=[1j*logical_x[temp]@logical_z[temp] for temp in range(len(logical_x))]  # 粒子数算符组作为逻辑算符组

    ##  获取数目
    majorana_number=code.physical_number
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    pauli_number=stabilizer_number

    ##  生成初始态
    majorana_state=stabilizers_x.tolist()+stabilizers_z.tolist()+logical_occupy+[MajoranaOperator([], [], 1) for _ in range(pauli_number)]
    pauli_state=[PauliOperator([],[],1) for _ in range(majorana_number)]+[PauliOperator([],[temp],1) for temp in range(pauli_number)]

    ##  ---生成线路---
    ##  强制初始化
    circuit = Circuit()
    circuit.append({"name":"FORCE",
                      "majorana_state":majorana_state,
                      'pauli_state':pauli_state,
                      'pauli_number':pauli_number,
                      'majorana_number':majorana_number})

    ##  第一轮测量假设完美的初始化
    observable_include= []  # 记录可观测量的索引
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        observable_include.append(len(circuit.measurements)-1)
    for i,stabilizer in enumerate(stabilizers_x):
        circuit.append({"name":"MPP","target":stabilizer})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit.append({"name":"MPP","target":stabilizer})

    ##  添加第一轮噪声
    for i in range(majorana_number):
        circuit.append({"name": "FDEPOLARIZE1", "target": i, "p": p_noise})
    for i in range(pauli_number):
        circuit.append({"name": "DEPOLARIZE1", "target": i, "p": p_noise})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  添加稳定子测量
        for i,stabilizer in enumerate(stabilizers_x):
            sequence_temp=syndrome_majorana_css_measurement_circuit(stabilizer, i, 'x', p_noise, p_measure)
            for temp in sequence_temp:
                circuit.append(temp)
        for i,stabilizer in enumerate(stabilizers_z):
            sequence_temp=syndrome_majorana_css_measurement_circuit(stabilizer, i + len(stabilizers_x), 'z', p_noise, p_measure)
            for temp in sequence_temp:
                circuit.append(temp)

        ##  添加检测器
        for i in range(stabilizer_number):
            circuit.append({"name":"DETECTOR","target":[-i - 1, -i - stabilizer_number-1]})

    ##  最后一轮测量假设是没有噪声的
    ##  添加稳定子测量
    for i,stabilizer in enumerate(stabilizers_x):
        circuit.append({"name":"MPP","target":stabilizer})
    for i,stabilizer in enumerate(stabilizers_z):
        circuit.append({"name":"MPP","target":stabilizer})

    ##  添加检测器
    for i in range(stabilizer_number):
        circuit.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  测量逻辑算符
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        circuit.append({"name":"OBSERVABLE_INCLUDE","target":[len(circuit.measurements) - 1, observable_include[i]]})

    ##  ---返回线路---
    return circuit


#%%  KEY：将Pauli CSS code转换为现象级噪声下的测试线路
def PauliCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number:int):
    pass


#%%  KEY：===将Pauli CSS code转换为电路级噪声下的测试线路===
def PauliCSSCode2CircuitLevelCircuit(code,p_noise,p_measure,cycle_number:int):
    ##  ---数据预处理---
    ##  获取稳定子
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z

    ##  获取逻辑算符
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z

    ##  获取数目
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    data_number=code.physical_number
    pauli_number=data_number+stabilizer_number

    ##  生成初始态
    ##  添加稳定子态
    pauli_state = []
    for i in range(len(stabilizers_x)):
        pauli_state.append(stabilizers_x[i])
    for i in range(len(stabilizers_z)):
        pauli_state.append(stabilizers_z[i])
    pauli_state_x=copy.deepcopy(pauli_state)
    pauli_state_z=copy.deepcopy(pauli_state)

    ##  添加逻辑态
    for i in range(len(logical_x)):
        pauli_state_x.append(logical_x[i])
    for i in range(len(logical_z)):
        pauli_state_z.append(logical_z[i])

    ##  添加辅助qubit态
    for i in range(data_number,stabilizer_number+data_number):
        pauli_state_x.append(PauliOperator([],[i],1))
        pauli_state_z.append(PauliOperator([],[i],1))

    ##  ---生成线路---
    ##  强制初始化
    circuit_x = Circuit()
    circuit_x.append({"name": "FORCE",
                      "majorana_state": [MajoranaOperator([],[],1)]*len(pauli_state_x),
                      'pauli_state': pauli_state_x,
                      'pauli_number': pauli_number,
                      'majorana_number': 0})
    circuit_z = Circuit()
    circuit_z.append({"name": "FORCE",
                      "majorana_state": [MajoranaOperator([],[],1)]*len(pauli_state_z),
                      'pauli_state': pauli_state_z,
                      'pauli_number': pauli_number,
                      'majorana_number': 0})

    ##  第一轮测量假设完美的初始化
    observable_include = []  # 记录可观测量的索引
    for i in range(len(logical_x)):
        circuit_x.append({"name": "MPP", "target": logical_x[i]})
        circuit_z.append({"name": "MPP", "target": logical_z[i]})
        observable_include.append(len(circuit_x.measurements) - 1)
    for i, stabilizer in enumerate(stabilizers_x):
        circuit_x.append({"name": "MPP", "target": stabilizer})
        circuit_z.append({"name": "MPP", "target": stabilizer})
    for i, stabilizer in enumerate(stabilizers_z):
        circuit_z.append({"name": "MPP", "target": stabilizer})
        circuit_x.append({"name": "MPP", "target": stabilizer})

    ##  施加第一轮噪声
    for i in range(pauli_number):
        circuit_z.append({"name": "DEPOLARIZE1", "target": i, "p": p_noise})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  测量稳定子
        for i, stabilizer in enumerate(stabilizers_x):
            sequence_temp = syndrome_pauli_css_measurement_circuit(stabilizer, i + data_number, 'x', p_noise, p_measure)
            for temp in sequence_temp:
                circuit_z.append(temp)
                circuit_x.append(temp)
        for i, stabilizer in enumerate(stabilizers_z):
            sequence_temp = syndrome_pauli_css_measurement_circuit(stabilizer, i + data_number + len(stabilizers_x), 'z', p_noise, p_measure)
            for temp in sequence_temp:
                circuit_z.append(temp)
                circuit_x.append(temp)

        ##  添加检测器
        for i in range(stabilizer_number):
            circuit_z.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})
            circuit_x.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  最后一轮测量假设是没有噪声的
    ##  测量稳定子
    for i, stabilizer in enumerate(stabilizers_x):
        circuit_z.append({"name": "MPP", "target": stabilizer})
        circuit_x.append({"name": "MPP", "target": stabilizer})
    for i, stabilizer in enumerate(stabilizers_z):
        circuit_z.append({"name": "MPP", "target": stabilizer})
        circuit_x.append({"name": "MPP", "target": stabilizer})

    ##  添加检测器
    for i in range(stabilizer_number):
        circuit_z.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})
        circuit_x.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  测量逻辑算符
    for i in range(len(logical_x)):
        circuit_z.append({"name": "MPP", "target": logical_z[i]})
        circuit_x.append({"name": "MPP", "target": logical_x[i]})
        circuit_z.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_z.measurements) - 1, observable_include[i]]})
        circuit_x.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_x.measurements) - 1, observable_include[i]]})

    ##  ---返回线路---
    return circuit_x, circuit_z


#%%  TODO：===将Majorana CSS code转换为码能力下的测试线路===
def MajoranaCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number:int):
    pass


#%%  TODO：===将Pauli CSS code转换为码能力下的测试线路===
def PauliCSSCode2CodeCapacityCircuit(code,p_noise,p_measure,cycle_number:int):
    pass


# %%  KEY：===生成Majorana CSS stabilizer测量线路===
def syndrome_majorana_css_measurement_circuit(stabilizer, qubit_index, type, p_noise, p_measure):

    ##  ———数据预处理———
    sequence = []  # 线路序列
    flag = True  # 门类型标志

    ##  将qubit置于负号匹配
    sequence.append({'name': 'X', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})

    ##  判断稳定子类型
    if type == 'x' or type == 'X':
        occupy=stabilizer.occupy_x
    elif type == 'z' or type == 'Z':
        occupy=stabilizer.occupy_z
    else:
        raise ValueError

    ##  ---生成纠缠线路---
    ##  生成前一半线路
    for j in range(len(occupy)):
        majorana_index_now = occupy[j]

        ##  最后一位与qubit作用CNX gate
        if j == len(occupy) - 1:
            sequence.append({'name': 'CNX', 'target': [majorana_index_now, qubit_index], })
            sequence.append({'name': 'FDEPOLARIZE1', 'target': majorana_index_now, 'p': p_noise})
            sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})
            break

        majorana_index_down = occupy[j + 1]  # 后一个fermionic site
        ##  作用braid gate
        if flag:
            ##  根据稳定子类型选择braid形式
            if type == 'X' or type == 'x':
                order_target = [majorana_index_down, majorana_index_now]
            elif type == 'Z' or type == 'z':
                order_target = [majorana_index_now, majorana_index_down]
            else:
                raise ValueError

            ##  添加braid gate
            sequence.append({"name": "B", "target": order_target, })
            sequence.append({'name': 'FDEPOLARIZE1', 'target': order_target, 'p': p_noise})
            flag = False

        ##  作用CNN gate
        else:
            sequence.append({'name': 'CNN', 'target': [majorana_index_now, majorana_index_down], })
            sequence.append({'name': 'FDEPOLARIZE1', 'target': [majorana_index_now, majorana_index_down], 'p': p_noise})
            flag = True

    ##  生成syndrome extraction circuit的另一半
    flag = True
    for j in range(len(occupy) - 1):
        majorana_index_now = occupy[-1 - j]  # 当前的fermionic site
        majorana_index_up = occupy[-1 - j - 1]  # 上一个fermionic site

        ##  作用braid gate
        if flag:
            if type == 'X' or type == 'x':
                order_target = [majorana_index_now, majorana_index_up]
            elif type == 'Z' or type == 'z':
                order_target = [majorana_index_up, majorana_index_now]
            else:
                raise ValueError
            sequence.append({'name': 'N', 'target': [majorana_index_now]})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': majorana_index_now, 'p': p_noise})
            sequence.append({'name': 'braid', 'target': order_target})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': order_target, 'p': p_noise})
            sequence.append({'name': 'N', 'target': [majorana_index_now]})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': majorana_index_now, 'p': p_noise})
            flag = False

        ##  作用CNN gate
        else:
            sequence.append({'name': 'CNN', 'target': [majorana_index_now, majorana_index_up]})
            sequence.append({'name': 'FDEPOLARIZE1', 'target': [majorana_index_now, majorana_index_up], 'p': p_noise})
            flag = True

    ##  在qubit上测量结果并重置
    sequence.append({'name': 'MZ', 'target': qubit_index, 'p': p_measure})
    sequence.append({'name': 'R', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})

    ##  ———返回线路序列———
    return sequence


#%%  KEY：===生成Pauli CSS stabilizer测量线路===
def syndrome_pauli_css_measurement_circuit(stabilizer, qubit_index, type, p_noise, p_measure):

    ##  ———数据预处理———
    sequence = []  # 线路序列

    ##  判断稳定子类型
    if type == 'x' or type == 'X':
        occupy=stabilizer.occupy_x
    elif type == 'z' or type == 'Z':
        occupy=stabilizer.occupy_z
    else:
        raise ValueError

    ##  ---生成纠缠线路---
    for j in range(len(occupy)):
        if type == 'Z' or type == 'z':
            sequence.append({'name': 'CX', 'target': [occupy[j], qubit_index]})
            sequence.append({'name': 'DEPOLARIZE1', 'target': [occupy[j], qubit_index], 'p': p_noise})
        elif type == 'X' or type == 'x':
            sequence.append({'name': 'H', 'target': occupy[j]})
            sequence.append({'name': 'DEPOLARIZE1', 'target': occupy[j], 'p': p_noise})
            sequence.append({'name': 'CX', 'target': [occupy[j], qubit_index]})
            sequence.append({'name': 'DEPOLARIZE1', 'target': [occupy[j],qubit_index], 'p': p_noise})
            sequence.append({'name': 'H', 'target': occupy[j]})
            sequence.append({'name': 'DEPOLARIZE1', 'target': occupy[j], 'p': p_noise})
        else:
            raise ValueError

    ##  在qubit上测量结果并重置
    sequence.append({'name': 'MZ', 'target': qubit_index, 'p': p_measure})
    sequence.append({'name': 'R', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': p_noise})

    ##  ———返回线路序列———
    return sequence



import copy
from abc import abstractmethod, ABC
import numpy as np
from extendedstim.Math.BinaryArray import BinaryArray as ba
from extendedstim.tools import isinteger, islist


class Operator(ABC):

    #%%  USER：===构造方法===
    def __init__(self, occupy_x, occupy_z, coff):
        assert islist(occupy_x)
        assert islist(occupy_z)
        occupy_x=copy.deepcopy(occupy_x)
        occupy_z=copy.deepcopy(occupy_z)
        self.occupy_x = np.array(occupy_x, dtype=int)
        self.occupy_z = np.array(occupy_z, dtype=int)
        self.occupy_x.sort()
        self.occupy_z.sort()
        self.coff = coff

    #%%  USER：===重载运算符===
    ##  USER：---矩阵乘法---
    @abstractmethod
    def __matmul__(self, other):
        pass

    ##  USER：---右矩阵乘法---
    @abstractmethod
    def __rmatmul__(self, other):
        pass

    ##  USER：---左标量乘法---
    @abstractmethod
    def __mul__(self, other):
        pass

    ##  USER：---右标量乘法---
    @abstractmethod
    def __rmul__(self, other):
        pass

    ##  USER：---字符串表示---
    @abstractmethod
    def __str__(self):
        pass

    ##  USER：---相等判断---
    @abstractmethod
    def __eq__(self, other):
        pass

    ##  USER：---取负---
    @abstractmethod
    def __neg__(self):
        pass

    #%%  USER：属性方法
    ##  USER：---算符的权重---
    @property
    def weight(self)->int:
        return len(self.occupy_x) + len(self.occupy_z)

    ##  USER：---判断算符是否是厄米算符---
    @property
    @abstractmethod
    def is_hermitian(self)->bool:
        pass

    ##  USER：---求算符的对偶算符---
    @property
    @abstractmethod
    def dual(self):
        pass

    #%%  USER：===对象方法===
    ##  USER：---更换索引方式---
    @abstractmethod
    def index_map(self, index):
        assert isinstance(index, np.ndarray) or isinstance(index, list)
        assert len(index) >= max(self.occupy_x.max(), self.occupy_z.max()) + 1
        x = np.array([index[i] for i in self.occupy_x], dtype=int)
        z = np.array([index[i] for i in self.occupy_z], dtype=int)
        return x, z, self.coff

    ##  USER：---获取向量表示---
    def get_vector(self, number):
        assert isinteger(number)
        vector = ba.zeros(number * 2)
        vector[self.occupy_x * 2] = 1
        vector[self.occupy_z * 2 + 1] = 1
        return vector

    ##  USER：---检查是否存在量子位上的算符占据---
    def is_exist_occupy_x(self, index):
        assert isinteger(index)
        occupy = np.where(self.occupy_x == index)[0]
        if len(occupy) == 0:
            return None
        return occupy[0]

    ##  USER：---检查是否存在量子位上的算符占据---
    def is_exist_occupy_z(self, index):
        assert isinteger(index)
        occupy = np.where(self.occupy_z == index)[0]
        if len(occupy) == 0:
            return None
        return occupy[0]

    ##  USER：---复制算符---
    @abstractmethod
    def copy(self):
        pass

    ##  USER：---将算符在index位置切分---
    @abstractmethod
    def split(self, index):
        assert isinteger(index)
        left_x=[temp for temp in self.occupy_x if temp < index]
        left_z=[temp for temp in self.occupy_z if temp < index]
        right_x=[temp for temp in self.occupy_x if temp > index]
        right_z=[temp for temp in self.occupy_z if temp > index]
        if self.is_exist_occupy_x(index) is not None and self.is_exist_occupy_z(index) is not None:
            middle_x=[index]
            middle_z=[index]
        elif self.is_exist_occupy_x(index) is not None:
            middle_x=[index]
            middle_z=[]
        elif self.is_exist_occupy_z(index) is not None:
            middle_x=[]
            middle_z=[index]
        else:
            middle_x=[]
            middle_z=[]
        return left_x, left_z, middle_x, middle_z, right_x, right_z

    #%%  USER：静态方法
    ##  USER：---求多个算符的矩阵表示---
    @staticmethod
    def get_matrix(ops, number):
        assert islist(ops)
        matrix = None
        for i, op in enumerate(ops):
            vector = op.get_vector(number)
            if i == 0:
                matrix = vector
            else:
                matrix = ba.vstack(matrix, vector)
        return matrix

    # %%  USER：静态方法
    ##  USER：---定义一个厄米算符，从占据处表示---
    @staticmethod
    @abstractmethod
    def HermitianOperatorFromOccupy(occupy_x, occupy_z):
        pass

    ##  USER：---定义一个厄米算符，从向量表示---
    @staticmethod
    @abstractmethod
    def HermitianOperatorFromVector(vector):
        pass

    ##  USER：---检查两个算符是否对易---
    @staticmethod
    @abstractmethod
    def commute(A,B):
        pass



from collections.abc import Iterable
import copy
import qiskit
import numpy as np
import stim
import stimbposd
from qiskit.circuit import CircuitError
from qiskit.circuit.library import XGate, ZGate
# noinspection PyUnresolvedReferences
from tesseract_decoder import tesseract
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator
from extendedstim.Platform.Platform import Platform
from extendedstim.tools import isinteger, islist


class Circuit:
    #%%  USER：===构造方法===
    def __init__(self):
        self.majorana_number = 0  # fermionic sites的数目
        self.pauli_number = 0  # qubits的数目
        self._sequence = []  # 量子线路的操作序列（真实计算使用）
        self.sequence = []  # 量子线路的操作序列
        self.noise = []  # 量子线路中噪声的索引
        self.measurements = []  # 量子线路中测量的索引
        self.detectors = []  # 量子线路中测量的结果探测器对self.measurements的索引
        self.observables = []  # 量子线路中可观测量的对self.measurements的索引
        self._dem=None  # 量子线路错误模型
        self._dem_str_list=[]  # 量子线路错误模型的字符串表示

    #%%  USER：===重载运算符===
    ##  USER：---获取序列中的元素---
    def __getitem__(self, item):
        return self._sequence[item]

    ##  USER：---设置序列中的元素---
    def __setitem__(self, key, value):
        self._sequence[key] = value

    #%%  USER：===对象方法===
    ##  USER：---添加量子线路组分操作---
    def append(self, params):
        name=params["name"]
        example={
            'name':'name',  # 线路操作的名称
            'target':[1,2],  # 操作作用的对象
            'p':0.1,  # 操作对应的概率，不一定有
            'index':2,  # 测量对应平台上的stabilizers的序号
            'majorana_state':[MajoranaOperator([],[],1)],  # 强制初始化的majorana state
            'pauli_state':[PauliOperator([],[],1)],  # 强制初始化的pauli state
            'pauli_number':10,  # 强制初始化的qubits数目
            'majorana_number':10,  # 强制初始化的fermionic sites数目
        }

        ##  添加single-qubit gate
        if name == 'X' or name == 'Y' or name == 'Z' or name == 'H' or name == 'S' or name=='P':
            if isinteger(params['target']):
                self._sequence.append(params.copy())
                self.sequence.append(self._sequence[-1])
            elif islist(params['target']):
                for temp in params['target']:
                    params_temp={'name':name, 'target':temp}
                    self._sequence.append(params_temp)
                    self.sequence.append(self._sequence[-1])

        ##  添加single-qubit上的噪声
        elif name == 'X_ERROR' or name == 'Y_ERROR' or name == 'Z_ERROR':
            if isinteger(params['target']):
                self._sequence.append(params.copy())
                self.sequence.append(self._sequence[-1])
            elif islist(params['target']):
                for temp in params['target']:
                    params_temp={'name':name, 'target':temp}
                    self._sequence.append(params_temp)
                    self.sequence.append(self._sequence[-1])

        ##  添加single-fermionic-site gate
        elif name == 'U' or name == 'V' or name == 'N' or name == 'P':
            if isinteger(params['target']):
                self._sequence.append(params.copy())
                self.sequence.append(self._sequence[-1])
            elif islist(params['target']):
                for temp in params['target']:
                    params_temp={'name':name, 'target':temp}
                    self._sequence.append(params_temp)
                    self.sequence.append(self._sequence[-1])

        ##  添加single-fermionic-site上的噪声
        elif name == 'U_ERROR' or name == 'V_ERROR' or name == 'N_ERROR':
            if isinteger(params['target']):
                self._sequence.append(params.copy())
                self.sequence.append(self._sequence[-1])
            elif islist(params['target']):
                for temp in params['target']:
                    params_temp={'name':name, 'target':temp}
                    self._sequence.append(params_temp)
                    self.sequence.append(self._sequence[-1])

        ##  添加受控非门
        elif name == 'CX' or name == 'CNX' or name=='CNN':
            self._sequence.append(params.copy())
            self.sequence.append(self._sequence[-1])

        ##  添加交错门
        elif name=='B' or name=='braid':
            x,y=params['target']
            if x==y and 'verse' not in params:
                self._sequence.append({'name': 'P', 'target':x})
                self.sequence.append(self._sequence[-1])
            elif x==y and 'verse' in params and params['verse']==True:
                self._sequence.append({'name': 'P', 'target':y})
                self.sequence.append(self._sequence[-1])
                self._sequence.append({'name': 'N', 'target': y})
                self.sequence.append(self._sequence[-1])
            elif x!=y:
                self._sequence.append({'name': 'braid', 'target':[x, y]})
                self.sequence.append(self._sequence[-1])

        ##  添加single qubit and single fermionic site上的测量
        elif name == 'MZ' or name == 'MN':
            if islist(params['target']):
                for temp in params['target']:
                    dict_temp = copy.deepcopy(params)
                    dict_temp['target']=temp
                    self.append(dict_temp)
            elif isinteger(params['target']):
                dict_temp = copy.deepcopy(params)
                dict_temp['name'] = 'MPP'
                if name == 'MZ':
                    dict_temp['target']=PauliOperator([], [dict_temp["target"]], 1)
                elif name == 'MN':
                    dict_temp['target']=MajoranaOperator([dict_temp["target"]], [dict_temp["target"]], 1j)
                self.append(dict_temp)

        ##  添加qubit上的去极化噪声
        elif name == 'DEPOLARIZE1':
            assert 'p' in params
            if isinteger(params['target']):
                self.sequence.append({'name': 'DEPOLARIZE1', 'target':params['target'],'p':params['p']})
                fix = (1-np.sqrt(1-4*params["p"]/3))/2
                self._sequence.append({'name': 'X_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence) - 1)
                self._sequence.append({'name': 'Y_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence) - 1)
                self._sequence.append({'name': 'Z_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence) - 1)
            elif islist(params['target']):
                for temp in params['target']:
                    self.append({'name': params['name'], 'target': temp, 'p': params["p"]})

        ##  添加fermionic site上的去极化噪声
        elif name == 'FDEPOLARIZE1':
            assert 'p' in params
            if isinteger(params['target']):
                self.sequence.append({'name':'FDEPOLARIZE1', 'target':params['target'],'p':params['p']})
                fix = (1-np.sqrt(1-4*params["p"]/3))/2
                self._sequence.append({'name': 'U_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence) - 1)
                self._sequence.append({'name': 'V_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence) - 1)
                self._sequence.append({'name': 'N_ERROR', 'target': params["target"], 'p': fix})
                self.noise.append(len(self._sequence) - 1)
            elif islist(params['target']):
                for temp in params['target']:
                    self.append({'name': params['name'], 'target': temp, 'p': params["p"]})

        ##  强制初始化
        elif name=='FORCE':
            assert 'majorana_state' in params
            assert 'pauli_state' in params
            assert 'pauli_number' in params
            assert 'majorana_number' in params
            stabilizers_majorana = copy.deepcopy(params["majorana_state"])
            stabilizers_pauli = copy.deepcopy(params["pauli_state"])
            self.pauli_number=params["pauli_number"]
            self.majorana_number=params["majorana_number"]
            self._sequence.append({'name': 'FORCE', 'majorana_state': stabilizers_majorana, 'pauli_state': stabilizers_pauli})

        ##  添加string算符的测量
        elif name == 'MPP':
            assert 'target' in params
            if islist(params['target']):
                for i,temp in enumerate(params['target']):
                    dict_temp = copy.deepcopy(params)
                    dict_temp['target']=temp
                    if 'index' in params:
                        dict_temp['index']=params['index'][i]
                    self.append(dict_temp)
            elif isinstance(params['target'], (PauliOperator, MajoranaOperator)):
                dict_temp = {'name': 'MPP', 'target': params['target']}
                if 'index' in params:
                    dict_temp['index']=params['index']
                self._sequence.append(dict_temp)
                self.measurements.append(len(self._sequence) - 1)
                if 'p' in params:
                    self._sequence.append({'name': 'M_ERROR', 'p': params["p"]})
                    self.noise.append(len(self._sequence) - 1)
                    self.sequence.append({'name': 'MPP', 'target': params['target'], 'p': params["p"]})
                else:
                    self.sequence.append(self._sequence[-1])
            else:
                raise ValueError

        ##  添加监视器
        elif name == 'DETECTOR':
            assert 'target' in params
            target=params['target']
            if all(target[i] < 0 for i in range(len(target))):
                together = [len(self.measurements) + temp for temp in target]  # 在测量中找到对应索引
                self.detectors.append(together)
                self.sequence.append({'name': 'DETECTOR', 'target': target})
            elif all(target[i] >= 0 for i in range(len(target))):
                self.detectors.append([temp for temp in target])
                self.sequence.append({'name': 'DETECTOR', 'target': [-len(self.measurements)+temp for temp in target]})
            else:
                raise ValueError("DETECTOR gate target must be consecutive")

        ##  添加可观测量
        elif name == 'OBSERVABLE_INCLUDE':
            assert 'target' in params
            target=params['target']
            if all(target[i] < 0 for i in range(len(target))):
                together = [len(self.measurements) + temp for temp in target]  # 在测量中找到索引
                self.observables.append(together)
                self.sequence.append({'name': 'OBSERVABLE_INCLUDE', 'target': target})
            elif all(target[i] >= 0 for i in range(len(target))):
                together = [temp for temp in target]  # 在测量中找到索引
                self.observables.append(together)
                self.sequence.append({'name': 'OBSERVABLE_INCLUDE', 'target': [-len(self.measurements) + temp for temp in target]})
            else:
                raise ValueError("OBSERVABLE_INCLUDE gate target must be consecutive")

        ##  添加qubit重置
        elif name == 'R':
            assert 'target' in params
            target=params['target']
            if isinteger(target):
                self._sequence.append({'name': name, 'target': target})
                self.sequence.append(self._sequence[-1])
                if target == self.pauli_number:
                    self.pauli_number += 1
                elif target > self.pauli_number or target < 0:
                    raise ValueError("R gate target must be consecutive")
                else:
                    pass
            elif islist(target):
                for temp in target:
                    self.append({'name': name, 'target': temp})
            else:
                raise ValueError

        ##  添加fermionic site重置
        elif name == 'FR':
            assert 'target' in params
            target=params['target']
            if isinteger(target):
                self._sequence.append({'name': name, 'target': target})
                self.sequence.append(self._sequence[-1])
                if target == self.majorana_number :
                    self.majorana_number = target+1
                elif target > self.majorana_number or target < 0:
                    raise ValueError("FR gate target must be consecutive")
                else:
                    pass
            elif islist(target):
                for temp in target:
                    self.append({'name': name, 'target': temp})
            else:
                raise ValueError
        else:
            raise NotImplementedError

    ##  USER：---生成无噪声的线路---
    def ideal_circuit(self):
        sequence = copy.deepcopy(self._sequence)
        for i in range(len(self.noise)):
            gate = sequence[self.noise[i]]
            assert isinstance(gate, dict)
            gate['p'] = 0
        ideal_circuit = Circuit()
        ideal_circuit.majorana_number = self.majorana_number
        ideal_circuit.pauli_number = self.pauli_number
        ideal_circuit._sequence = sequence
        ideal_circuit.measurements = self.measurements
        ideal_circuit.detectors = self.detectors
        ideal_circuit.observables = self.observables
        ideal_circuit.noise = self.noise
        return ideal_circuit

    ##  USER：---执行线路并返回测量结果---
    def execute(self):
        platform = Platform()  # 生成量子平台
        platform.initialize(self.majorana_number, self.pauli_number)  # 定义量子平台的qubits和fermionic sites数目
        measurement_sample = np.empty(len(self.measurements), dtype=int)  # 生成测量值的样本数组
        flag_measurement = 0

        ##  遍历整个操作序列
        for i, gate in enumerate(self._sequence):
            name = gate['name']
            p = None
            target = None
            if 'target' in gate:
                target = gate['target']
            if 'p' in gate:
                p = gate['p']

            ##  执行单门
            if name in ['X','Y','Z','H','S','U','V','N']:
                assert target is not None
                if isinstance(target, int):
                    eval('platform.'+name+'(target)')
                elif isinstance(target,Iterable):
                    for qubit in target:
                        eval('platform.'+name+f'({qubit})')

            ##  执行双门
            elif name in ['CX','CNX','braid','CNN']:
                assert target is not None
                if name=='braid':
                    platform.braid(target[0],target[1])
                if name=='CNN':
                    platform.CNN(target[0],target[1])
                if name=='CNX':
                    platform.CNX(target[0],target[1])
                if name=='CX':
                    platform.CX(target[0],target[1])

            ##  执行重置
            elif name == 'R':
                assert target is not None
                if isinstance(target, int):
                    platform.reset(target)
                elif isinstance(target,Iterable):
                    for qubit in target:
                        platform.reset(qubit)
            elif name == 'FR':
                assert target is not None
                if isinstance(target, int):
                    platform.fermionic_reset(target)
                elif isinstance(target,Iterable):
                    for qubit in target:
                        platform.fermionic_reset(qubit)

            ##  执行误差门
            elif name in ['X_ERROR','Y_ERROR','Z_ERROR','U_ERROR','V_ERROR','N_ERROR']:
                assert target is not None
                assert p is not None
                if isinstance(target, int):
                    eval('platform.'+name.lower()+f'({target},{p})')
                elif isinstance(target,Iterable):
                    for qubit in target:
                        eval('platform.'+name.lower()+f'({qubit},{p})')
            elif name == 'M_ERROR':
                assert p is not None
                if np.random.rand() < p:
                    measurement_sample[flag_measurement-1] = -measurement_sample[flag_measurement-1]

            ##  执行测量
            elif name == 'MPP':
                assert target is not None
                if isinstance(target, Iterable):
                    for op in target:
                        if 'index' in gate:
                            measurement_sample[flag_measurement] = platform.measure(op, gate['index'])
                        else:
                            measurement_sample[flag_measurement] = platform.measure(op)
                        flag_measurement += 1
                elif isinstance(target,PauliOperator) or isinstance(target,MajoranaOperator):
                    if 'index' in gate:
                        measurement_sample[flag_measurement] = platform.measure(target, gate['index'])
                    else:
                        measurement_sample[flag_measurement] = platform.measure(target)
                    flag_measurement += 1
                else:
                    raise ValueError("M_ERROR gate target must be consecutive")

            ##  执行强制初始化
            elif name=='FORCE':
                assert gate['majorana_state'] is not None
                assert gate['pauli_state'] is not None
                platform.initialize(self.majorana_number, self.pauli_number,gate['majorana_state'],gate['pauli_state'])

            else:
                raise NotImplementedError

        ##  计算探测器的结果
        detector_sample = np.empty(len(self.detectors), dtype=bool)
        flag_detector = 0
        for i, detector in enumerate(self.detectors):
            value = measurement_sample[detector][0]
            detector_sample[flag_detector] = False
            for temp in measurement_sample[detector]:
                if value == temp:
                    continue
                else:
                    detector_sample[flag_detector] = True
                    break
            flag_detector += 1

        observable_sample = np.empty(len(self.observables), dtype=bool)
        flag_observable = 0
        for i, observable in enumerate(self.observables):
            if len(observable) == 1:
                if measurement_sample[observable][0] == 1:
                    observable_sample[flag_observable] = False
                else:
                    observable_sample[flag_observable] = True
            else:
                value = measurement_sample[observable][0]
                observable_sample[flag_observable] = False
                for temp in measurement_sample[observable]:
                    if value == temp:
                        continue
                    else:
                        observable_sample[flag_observable] = True
                        break
            flag_observable += 1
        return measurement_sample, detector_sample, observable_sample

    ##  USER：--修改error model的错误几率--
    def noise_amplitude_fix(self,p_noise,p_measure):
        dem_str =''
        for i,temp in enumerate(self._dem_str_list):
            if temp[2]=='G_ERROR':
                temp[0]=f'error({p_noise/3}) '
            elif temp[2]=='M_ERROR':
                temp[0]=f'error({p_measure}) '
            dem_str += ('\n' + temp[0]+temp[1])
            self._dem=stim.DetectorErrorModel(dem_str)

    ##  USER：--生成检测错误模型--
    def detector_error_model(self) -> stim.DetectorErrorModel:
        if self._dem is not None:
            return self._dem
        ideal_circuit = self.ideal_circuit()
        measurement_sample_origin, detector_sample_origin, observable_sample_origin = ideal_circuit.execute()

        ##  执行检验线路的稳定性
        for time in range(5):
            measurement_sample, detector_sample, observable_sample= ideal_circuit.execute()
            assert np.all(detector_sample==detector_sample_origin),f'原始线路的detector不是稳定的'
            assert np.all(observable_sample==observable_sample_origin),f'原始线路的observable不是稳定的'

        errors = []
        dem_str = ''
        for i in range(len(ideal_circuit.noise)):
            print(i/len(ideal_circuit.noise))
            order = ideal_circuit.noise[i]
            gate_ideal = ideal_circuit._sequence[order]
            assert isinstance(gate_ideal, dict)
            gate_ideal['p'] = 1.1
            gate = self._sequence[order]
            assert isinstance(gate, dict)
            p = gate['p']
            measurement_sample, detector_sample, observable_sample = ideal_circuit.execute()
            detector_sample_diff = [detector_sample_origin[j] ^ detector_sample[j] for j in range(len(detector_sample))]
            observable_sample_diff = [observable_sample_origin[j] ^ observable_sample[j] for j in range(len(observable_sample))]
            errors.append(len(errors))
            detectors_trigger = np.where(np.array(detector_sample_diff) == True)[0]
            observables_trigger = np.where(np.array(observable_sample_diff) == True)[0]
            if len(detectors_trigger) > 0 or len(observables_trigger) > 0:
                temp_error=f'error({p}) '
                temp_trigger=''
                for index in detectors_trigger:
                    temp_trigger = temp_trigger + f' D{index}'
                for index in observables_trigger:
                    temp_trigger = temp_trigger + f' L{index}'
                temp = temp_error+temp_trigger
                if gate['name']=='M_ERROR':
                    self._dem_str_list.append([temp_error,temp_trigger,'M_ERROR'])
                else:
                    self._dem_str_list.append([temp_error,temp_trigger,'G_ERROR'])
                dem_str += ('\n' + temp)
            gate_ideal['p'] = 0
        dem = stim.DetectorErrorModel(dem_str)
        self._dem=dem
        return dem

    ##  USER：--生成解码函数--
    def decoder(self, method):
        dem = self.detector_error_model()
        if method == 'bposd':
            dec=stimbposd.bp_osd.BPOSD(model=dem,bp_method='min_sum',max_bp_iters=100)
        elif method=='tesseract':
            config=tesseract.TesseractConfig(dem=dem, det_beam=50)
            dec=config.compile_decoder()
        else:
            raise NotImplementedError
        return dec

    ##  USER：--执行线路并返回错误率--
    def sample(self,sample_number:int,method:str):
        dem= self.detector_error_model()
        sampler=dem.compile_sampler()
        decoder=self.decoder(method)
        detector_data, obs_data, error_data = sampler.sample(shots=sample_number)
        predictions = decoder.decode_batch(detector_data)
        num_errors = 0
        for shot in range(sample_number):
            actual_for_shot = obs_data[shot]
            predicted_for_shot = predictions[shot]
            if not np.array_equal(actual_for_shot, predicted_for_shot):
                num_errors += 1
        return num_errors/sample_number

    def stim_circuit(self):
        circuit=stim.Circuit()
        circuit.append('R',range(self.pauli_number))
        flag_measure=0
        for i in range(len(self.sequence)):
            gate=self.sequence[i]
            name=gate['name']
            if name == 'X' or name == 'Y' or name == 'Z' or name == 'H' or name == 'S' or name == 'P':
                circuit.append(name,[gate['target']])

            ##  添加single-qubit上的噪声
            elif name == 'X_ERROR' or name == 'Y_ERROR' or name == 'Z_ERROR':
                circuit.append(name, gate['target'],gate['p'])

            ##  添加single-fermionic-site gate
            elif name in['U','V','N','P','U_ERROR','V_ERROR','N_ERROR','CNX','CNN','B','braid','MN','FDEPOLARIZE1','FR']:
                raise NotImplementedError('stim只支持pauli circuit')

            ##  添加受控非门
            elif name == 'CX':
                target = gate['target']
                circuit.append(name, target)

            ##  添加qubit上的去极化噪声
            elif name == 'DEPOLARIZE1':
                target = gate['target']
                circuit.append(name, target,gate['p'])

            ##  强制初始化
            elif name == 'FORCE':
                raise NotImplementedError('stim不支持强制初始化')

            ##  添加string算符的测量
            elif name == 'MPP':
                op:PauliOperator=gate['target']
                occupy_x=op.occupy_x
                occupy_z=op.occupy_z
                if len(occupy_x)==0 and len(occupy_z)==1:
                    if 'p' in gate:
                        circuit.append('MZ', [occupy_z[0]],gate['p'])
                    else:
                        circuit.append('MZ', [occupy_z[0]])
                    continue
                elif len(occupy_x)==1 and len(occupy_z)==0:
                    if 'p' in gate:
                        circuit.append('MX', [occupy_x[0]],gate['p'])
                    else:
                        circuit.append('MX', [occupy_x[0]])
                    continue
                elif len(occupy_x)==1 and len(occupy_z)==1 and occupy_x[0]==occupy_z[0]:
                    if 'p' in gate:
                        circuit.append('MY', [occupy_z[0]],gate['p'])
                    else:
                        circuit.append('MY', [occupy_z[0]])
                    continue

                op_str=''
                for i in range(self.pauli_number):
                    if i in occupy_x and i in occupy_z:
                        op_str += 'Y'
                    elif i in occupy_z:
                        op_str += 'Z'
                    elif i in occupy_x:
                        op_str += 'X'
                    else:
                        op_str += '_'
                if 'p' in gate:
                    circuit.append('MPP', [stim.PauliString(op_str)],gate['p'])
                else:
                    circuit.append('MPP', [stim.PauliString(op_str)])

            ##  添加qubit重置
            elif name == 'R':
                circuit.append('R', [gate['target']])

            ##  检测器
            elif name == 'DETECTOR':
                circuit.append(name, [stim.target_rec(temp) for temp in gate['target']])

            ##  添加可观测量
            elif name == 'OBSERVABLE_INCLUDE':
                circuit.append(name, [stim.target_rec(temp) for temp in gate['target']],flag_measure)
                flag_measure+=1
            else:
                raise NotImplementedError
        return circuit

    ##  USER：--绘制线路图--
    def draw(self, filename):
        # 绘制一个带有barriers和更多寄存器中，绘制一个新的电路
        F = qiskit.QuantumRegister(self.majorana_number, name='F')
        Q = qiskit.QuantumRegister(self.pauli_number, name='Q')
        C = qiskit.ClassicalRegister(1, name='C')
        A = qiskit.QuantumRegister(1, name='A')
        circuit_qiskit = qiskit.QuantumCircuit(F, Q, C,A)
        braid = qiskit.circuit.ControlledGate(name='braid', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=XGate())
        cnn = qiskit.circuit.ControlledGate(name='CNN', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=ZGate())
        cnx = qiskit.circuit.ControlledGate(name='CNX', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=XGate())
        x_error = qiskit.circuit.Gate('X_ERROR', 1, label='X', params=[])
        y_error = qiskit.circuit.Gate('Y_ERROR', 1, label='Y', params=[])
        z_error = qiskit.circuit.Gate('Z_ERROR', 1, label='Z', params=[])
        u_error = qiskit.circuit.Gate('U_ERROR', 1, label='U', params=[])
        v_error = qiskit.circuit.Gate('V_ERROR', 1, label='V', params=[])
        n_error = qiskit.circuit.Gate('N_ERROR', 1, label='N', params=[])
        reset = qiskit.circuit.Gate('reset', 1, label=None, params=[])
        n=qiskit.circuit.Gate('N', 1, label='N', params=[])
        dep=qiskit.circuit.Gate('DEPOLARIZE', 1, label='D', params=[])

        for gate in self.sequence:
            if gate['name'] == 'R':
                circuit_qiskit.append(reset, [Q[gate['target']]])
            elif gate['name'] == 'FR':
                circuit_qiskit.append(reset, [F[gate['target']]])
            elif gate['name'] == 'X':
                circuit_qiskit.x(Q[gate['target']])
            elif gate['name'] == 'Y':
                circuit_qiskit.y(Q[gate['target']])
            elif gate['name'] == 'Z':
                circuit_qiskit.z(Q[gate['target']])
            elif gate['name'] == 'H':
                circuit_qiskit.h(Q[gate['target']])
            elif gate['name'] == 'S':
                circuit_qiskit.s(Q[gate['target']])
            elif gate['name']=='N':
                circuit_qiskit.append(n, [F[gate['target']]])
            elif gate['name'] == 'CX':
                circuit_qiskit.cx(Q[gate['target'][0]], Q[gate['target'][1]])
            elif gate['name'] == 'braid':
                circuit_qiskit.append(braid, [F[gate['target'][0]],F[gate['target'][1]]])
            elif gate['name'] == 'CNN':
                circuit_qiskit.append(cnn, [F[gate['target'][0]],F[gate['target'][1]]])
            elif gate['name'] == 'CNX':
                circuit_qiskit.append(cnx, [F[gate['target'][0]],Q[gate['target'][1]]])
            elif gate['name'] == 'X_ERROR':
                circuit_qiskit.append(x_error, [Q[gate['target']]])
            elif gate['name'] == 'Y_ERROR':
                circuit_qiskit.append(y_error, [Q[gate['target']]])
            elif gate['name'] == 'Z_ERROR':
                circuit_qiskit.append(z_error, [Q[gate['target']]])
            elif gate['name'] == 'U_ERROR':
                circuit_qiskit.append(u_error, [F[gate['target']]])
            elif gate['name'] == 'V_ERROR':
                circuit_qiskit.append(v_error, [F[gate['target']]])
            elif gate['name'] == 'N_ERROR':
                circuit_qiskit.append(n_error, [F[gate['target']]])
            elif gate['name'] == 'DEPOLARIZE1':
                circuit_qiskit.append(dep, [Q[gate['target']]])
            elif gate['name'] == 'FDEPOLARIZE1':
                circuit_qiskit.append(dep, [F[gate['target']]])
            elif gate['name'] == 'MPP':
                op = gate['target']
                if isinstance(op, MajoranaOperator):
                    f_flag_x = op.occupy_x
                    f_flag_z = op.occupy_z
                    f_flag_n = np.intersect1d(f_flag_x, f_flag_z)
                    f_flag_x = np.setdiff1d(f_flag_x, f_flag_n)
                    f_flag_z = np.setdiff1d(f_flag_z, f_flag_n)
                    f = np.concatenate([f_flag_x, f_flag_z, f_flag_n])
                    if len(f)>1:
                        mppx=qiskit.circuit.ControlledGate(name='MPPX', num_qubits=len(f)+1, params=[], label=None, num_ctrl_qubits=len(f), base_gate=XGate())
                        circuit_qiskit.append(mppx, F[f.tolist()]+[A[0]])
                        circuit_qiskit.measure(A[0], C[0])
                    else:
                        circuit_qiskit.measure(F[f[0]], C[0])
                elif isinstance(op, PauliOperator):
                    p_flag_x = op.occupy_x
                    p_flag_z = op.occupy_z
                    p_flag_y = np.intersect1d(p_flag_x, p_flag_z)
                    p_flag_x = np.setdiff1d(p_flag_x, p_flag_y)
                    p_flag_z = np.setdiff1d(p_flag_z, p_flag_y)
                    p = np.concatenate([p_flag_x, p_flag_y, p_flag_z])
                    if len(p)>1:
                        mppx=qiskit.circuit.ControlledGate(name='MPPX', num_qubits=len(p)+1, params=[], label=None, num_ctrl_qubits=len(p), base_gate=XGate())
                        circuit_qiskit.append(mppx, Q[p.tolist()]+[A[0]])
                        circuit_qiskit.measure(A[0], C[0])
                    else:
                        circuit_qiskit.measure(Q[p[0]], C[0])
                else:
                    raise CircuitError("cannot set parameters on immutable base gate")

        red='#E77081'
        blue='#5375CD'
        green='#00857B'
        grey='#8C92AC'
        purple='#5D548C'
        orange='#F15D22'
        pink='#FFACC5'
        cyan='#C9DCC4'
        circuit_qiskit.draw(output='mpl', filename=filename, style={
            'displaycolor': {'cx':None, 'cy':None, 'cz':None,
                             'X_ERROR': red, 'Y_ERROR': red, 'Z_ERROR': red,
                             'U_ERROR': red, 'V_ERROR': red, 'N_ERROR': red,
                             'R': cyan,'measure':grey,
                             'x':blue, 'y':blue, 'z':blue,'s':blue,'N':blue,'U':blue,'V':blue,
                             'CNN':blue,'CNX':purple,'braid':pink,
                             'MPPX':grey
                             },
            'fontsize': 12
        })

    ##  USER：复制函数
    def copy(self):
        return copy.deepcopy(self)


import numpy as np
from extendedstim.Physics.Operator import Operator


class MajoranaOperator(Operator):

    #%%  USER：===构造方法===
    def __init__(self, occupy_x, occupy_y, coff):
        super().__init__(occupy_x, occupy_y, coff)

    #%%  USER：===重载运算符===
    ##  USER：---矩阵乘法---
    def __matmul__(self, other):
        assert isinstance(other, MajoranaOperator)
        occupy_x = np.setxor1d(self.occupy_x, other.occupy_x, assume_unique=False)
        occupy_z = np.setxor1d(self.occupy_z, other.occupy_z, assume_unique=False)
        self_occupy = np.append(self.occupy_x * 2, self.occupy_z * 2 + 1)
        other_occupy = np.append(other.occupy_x * 2, other.occupy_z * 2 + 1)
        self_occupy = np.sort(self_occupy)
        other_occupy = np.sort(other_occupy)
        exchange_times = np.sum([np.count_nonzero(self_occupy > temp) for temp in other_occupy])
        if exchange_times % 2 == 1:
            factor = -1
        else:
            factor = 1
        return MajoranaOperator(occupy_x, occupy_z, self.coff * other.coff * factor)

    ##  USER：---右矩阵乘法---
    def __rmatmul__(self, other):
        assert isinstance(other, MajoranaOperator)
        return other.__matmul__(self)

    ##  USER：---标量乘法---
    def __mul__(self, other):
        assert other == 1 or other == -1 or other == 1j or other == -1j
        return MajoranaOperator(self.occupy_x, self.occupy_z, self.coff * other)

    ##  USER：---右标量乘法---
    def __rmul__(self, other):
        return self.__mul__(other)

    ##  USER：---字符串表示---
    def __str__(self):
        return "MajoranaOperator(occupy_x={},occupy_z={},coff={})".format(self.occupy_x, self.occupy_z, self.coff)

    ##  USER：---相等判断---
    def __eq__(self, other):
        assert isinstance(other, MajoranaOperator)
        return np.array_equal(self.occupy_x, other.occupy_x) and np.array_equal(self.occupy_z, other.occupy_z) and self.coff == other.coff

    ##  USER：---取负---
    def __neg__(self):
        return MajoranaOperator(self.occupy_x, self.occupy_z, -self.coff)

    # %%  USER：===属性方法===
    ##  USER：---算符是否是厄米算符---
    @property
    def is_hermitian(self):
        if np.mod(self.weight * (self.weight - 1) // 2,2) == 0:
            if not (self.coff == 1 or self.coff == -1):
                return False
        else:
            if not (self.coff == 1j or self.coff == -1j):
                return False
        return True

    ##  USER：---求算符的对偶算符---
    @property
    def dual(self):
        return MajoranaOperator(self.occupy_z, self.occupy_x, self.coff)

    # %%  USER：===对象方法===
    ##  USER：---将算符在index位置切分---
    def split(self, index):
        left_x, left_z, middle_x, middle_z, right_x, right_z = super().split(index)
        return MajoranaOperator(left_x, left_z, self.coff), MajoranaOperator(middle_x, middle_z, 1), MajoranaOperator(right_x, right_z, 1)

    ##  USER：---将算符映射成一个新的算符，修改index---
    def index_map(self, index):
        x, z, coff = super().index_map(index)
        return MajoranaOperator(x, z, coff)

    ##  USER：---复制方法---
    def copy(self):
        return MajoranaOperator(self.occupy_x.copy(), self.occupy_z.copy(), self.coff)

    # %%  USER：===静态方法===
    ##  USER：---定义一个厄米算符，从占据处表示---
    @staticmethod
    def HermitianOperatorFromOccupy(occupy_x, occupy_z):
        weight = len(occupy_x) + len(occupy_z)
        if (weight * (weight - 1) // 2) % 2 == 0:
            coff = 1
        else:
            coff = 1j
        return MajoranaOperator(occupy_x, occupy_z, coff)

    ##  USER：---定义一个厄米算符，从向量表示---
    @staticmethod
    def HermitianOperatorFromVector(vector):
        occupy_x = np.where(vector[0::2] == 1)[0]
        occupy_z = np.where(vector[1::2] == 1)[0]
        return MajoranaOperator.HermitianOperatorFromOccupy(occupy_x, occupy_z)

    ##  USER：---检查两个厄米算符是否对易---
    @staticmethod
    def commute(A, B):
        assert isinstance(A, MajoranaOperator) and isinstance(B, MajoranaOperator)
        overlap_x = len(np.intersect1d(A.occupy_x, B.occupy_x))
        overlap_z = len(np.intersect1d(A.occupy_z, B.occupy_z))
        weight = (len(A.occupy_x) + len(A.occupy_z)) * (len(B.occupy_x) + len(B.occupy_z))
        judge = overlap_x + overlap_z + weight
        return np.mod(judge, 2) == 0


import extendedstim.Code.QuantumCode.QuantumCode
import extendedstim.Code.QuantumCode.QuantumCSSCode
import extendedstim.Code.QuantumCode.PauliCode
import extendedstim.Code.QuantumCode.PauliCSSCode
import extendedstim.Code.QuantumCode.MajoranaCode
import extendedstim.Code.QuantumCode.MajoranaCSSCode
import extendedstim.Code.LinearCode.LinearCode
import extendedstim.Code.LinearCode.BicycleCode
import extendedstim.Code.LinearCode.FiniteEuclideanGeometryCode
import extendedstim.Code.LinearCode.FiniteProjectiveGeometryCode
import extendedstim.Circuit.Circuit
import extendedstim.Math.BinaryArray
import extendedstim.Circuit.Code2Circuit
import extendedstim.Circuit.Code2Circuit

PauliCode=extendedstim.Code.QuantumCode.PauliCode.PauliCode
PauliCSSCode=extendedstim.Code.QuantumCode.PauliCSSCode.PauliCSSCode
MajoranaCode=extendedstim.Code.QuantumCode.MajoranaCode.MajoranaCode
MajoranaCSSCode=extendedstim.Code.QuantumCode.MajoranaCSSCode.MajoranaCSSCode
LinearCode=extendedstim.Code.LinearCode.LinearCode.LinearCode
BicycleCode=extendedstim.Code.LinearCode.BicycleCode.BicycleCode
FiniteEuclideanGeometryCode=extendedstim.Code.LinearCode.FiniteEuclideanGeometryCode.FiniteEuclideanGeometryCode
FiniteProjectiveGeometryCode=extendedstim.Code.LinearCode.FiniteProjectiveGeometryCode.FiniteProjectiveGeometryCode
Code2Circuit=extendedstim.Circuit.Code2Circuit.Code2Circuit
Circuit=extendedstim.Circuit.Circuit.Circuit
BinaryArray=extendedstim.Math.BinaryArray.BinaryArray



import numpy as np
import galois
from itertools import product
from extendedstim.Code.LinearCode.LinearCode import LinearCode
from extendedstim.tools import isinteger


class FiniteProjectiveGeometryCode(LinearCode):

    #%%  USER：===构造方法===
    def __init__(self,s):
        ##  ---数据预处理---
        assert isinteger(s), "s必须是整数"

        ##  ---生成校验矩阵---
        self.s=s
        H_asym = generate_matrix(self.s)
        H_dual = np.hstack([H_asym[0:(H_asym.shape[0]-1)//2, 0:2 ** s - 1], H_asym[0:(H_asym.shape[0]-1)//2, 2 ** s:4 ** s+1]])
        assert np.all(H_dual@H_dual.T==0)
        super().__init__(H_dual)


#%%  KEY：生成Projective LDPC code
def generate_matrix(s):
    """构造非对称双包含量子 LDPC 码 C_asym"""
    # 步骤1: 构造射影平面
    points = construct_projective_plane(s)
    lines = get_lines(points)
    M_pi = construct_incidence_matrix(points, lines)
    n_points = len(points)

    # 步骤2: 构造超椭圆并分类点
    hyperoval, non_hyperoval = construct_hyperoval(points, s)
    non_hyperoval_indices = [points.index(p) for p in non_hyperoval]

    # 步骤3: 分类直线
    secant_lines, skew_lines = classify_lines(lines, hyperoval, points)

    # 步骤4: 构造 H_sk (skew lines)
    if skew_lines:
        H_sk = M_pi[skew_lines][:, non_hyperoval_indices]
        u_vector = np.ones((len(skew_lines), 1), dtype=int)
        H_sk = np.hstack([H_sk, u_vector])
    else:
        H_sk = np.zeros((0, len(non_hyperoval_indices) + 1), dtype=int)

    # 步骤5: 构造 H_se (secant lines)
    if secant_lines:
        H_se = M_pi[secant_lines][:, non_hyperoval_indices]
        u_vector = np.ones((len(secant_lines), 1), dtype=int)
        H_se = np.hstack([H_se, u_vector])
    else:
        H_se = np.zeros((0, len(non_hyperoval_indices) + 1), dtype=int)

    # 步骤6: 构造量子校验矩阵 (分块对角)
    n_sk = H_sk.shape[1] if H_sk.size > 0 else len(non_hyperoval_indices) + 1
    n_se = H_se.shape[1] if H_se.size > 0 else len(non_hyperoval_indices) + 1

    # 创建全零矩阵作为基础
    total_rows = H_sk.shape[0] + H_se.shape[0]
    total_cols = n_sk + n_se
    H_asym = np.zeros((total_rows, total_cols), dtype=int)

    # 填充 H_sk 部分
    if H_sk.size > 0:
        H_asym[:H_sk.shape[0], :H_sk.shape[1]] = H_sk

    # 填充 H_se 部分
    if H_se.size > 0:
        start_row = H_sk.shape[0]
        start_col = n_sk
        H_asym[start_row:start_row+H_se.shape[0], start_col:start_col+H_se.shape[1]] = H_se

    # 转换为伽罗瓦域矩阵
    GF2 = galois.GF(2)
    return GF2(H_asym)


def construct_projective_plane(s):
    """构造有限射影平面 PG(2, 2^s)"""
    q = 2 ** s
    GF = galois.GF(2 ** s)

    # 生成点集 [x, y, z] (齐次坐标)
    points = []
    for x, y, z in product(GF.elements, repeat=3):
        if x == 0 and y == 0 and z == 0:
            continue  # 排除原点

        # 标准化表示: 使第一个非零元素为1
        coords = [x, y, z]
        for i in range(3):
            if coords[i] != 0:
                scale = coords[i] ** -1
                point = tuple(c * scale for c in coords)
                if point not in points:
                    points.append(point)
                break
    return points


def get_lines(points):
    """获取所有直线"""
    lines = []
    for i in range(len(points)):
        for j in range(i + 1, len(points)):
            p1, p2 = points[i], points[j]
            # 计算通过两点的直线方程
            a = p1[1] * p2[2] - p1[2] * p2[1]
            b = p1[2] * p2[0] - p1[0] * p2[2]
            c = p1[0] * p2[1] - p1[1] * p2[0]

            # 标准化直线方程
            for k in range(3):
                if [a, b, c][k] != 0:
                    scale = [a, b, c][k] ** -1
                    a, b, c = a * scale, b * scale, c * scale
                    break
            line = (a, b, c)
            if line not in lines:
                lines.append(line)
    return lines

def construct_incidence_matrix(points, lines):
    """构造点线关联矩阵"""
    n_points = len(points)
    n_lines = len(lines)
    M = np.zeros((n_lines, n_points), dtype=int)

    for i, line in enumerate(lines):
        a, b, c = line
        for j, point in enumerate(points):
            x, y, z = point
            if a * x + b * y + c * z == 0:  # 点在直线上
                M[i, j] = 1
    return M


def construct_hyperoval(points, s):
    """构造超椭圆 (二次曲线 + 核点)"""
    q = 2 ** s
    GF = galois.GF(2 ** s)

    # 构造二次曲线 (x^2 = yz)
    conic_points = []
    for point in points:
        x, y, z = point
        if x ** 2 == y * z:  # 满足二次曲线方程
            conic_points.append(point)

    # 计算核点 (所有切线的交点)
    if conic_points:
        # 使用Galois域安全的加法
        x_sum = GF(0)
        y_sum = GF(0)
        z_sum = GF(0)
        for p in conic_points:
            x_sum += p[0]
            y_sum += p[1]
            z_sum += p[2]

        # 标准化核点
        nucleus = (x_sum, y_sum, z_sum)
        for i in range(3):
            if nucleus[i] != 0:
                scale = nucleus[i] ** -1
                nucleus = tuple(c * scale for c in nucleus)
                break
    else:
        # 默认核点 (当s=1时)
        nucleus = (GF(1), GF(0), GF(0))

    hyperoval = conic_points.copy()
    hyperoval.append(nucleus)
    non_hyperoval = [p for p in points if p not in hyperoval]

    return list(hyperoval), non_hyperoval

def classify_lines(lines, hyperoval, points):
    """分类直线: secant 或 skew"""
    secant_lines = []
    skew_lines = []

    for idx, line in enumerate(lines):
        # 计算直线与超椭圆的交点
        intersection_count = 0
        for point in hyperoval:
            x, y, z = point
            a, b, c = line
            if a * x + b * y + c * z == 0:  # 点在直线上
                intersection_count += 1

        if intersection_count == 2:  # secant line
            secant_lines.append(idx)
        elif intersection_count == 0:  # skew line
            skew_lines.append(idx)

    return secant_lines, skew_lines


import numpy as np


def isinteger(num):
    return isinstance(num, (int, float,np.int8,np.int16,np.int32, np.int64))


def islist(num):
    return isinstance(num, (list, tuple, np.ndarray))


def isfloat(num):
    return isinstance(num, (float, np.float32, np.float64,np.float128,np.float256))


def isreal(num):
    return isinteger(num) or isfloat(num)


def iscomplex(num):
    return isinteger(num) or isfloat(num) or isinstance(num, (complex,np.complex64, np.complex128,np.complex256,np.complex512))



import copy
from abc import abstractmethod
import numpy as np
from extendedstim.Code.QuantumCode.QuantumCode import QuantumCode
from extendedstim.Physics.Operator import Operator


class QuantumCSSCode(QuantumCode):

    #%%  USER：构造方法
    def __init__(self, generators_x, generators_z, physical_number):
        self.generators_x = np.array(generators_x,dtype=Operator)
        self.generators_z = np.array(generators_z,dtype=Operator)
        self.checker_number_x=len(generators_x)
        self.checker_number_z = len(generators_z)
        self._logical_operators_x=None
        self._logical_operators_z=None
        super().__init__(generators_x+generators_z, physical_number)

    #%%  USER：属性方法
    ##  USER：求逻辑算符（X方向）
    @property
    @abstractmethod
    def logical_operators_x(self):
        pass

    ##  USER：求逻辑算符（Z方向）
    @property
    @abstractmethod
    def logical_operators_z(self):
        pass

    ##  USER：求校验矩阵的秩（X方向）
    @property
    def rank_x(self):
        return self.check_matrix_x.rank

    ##  USER：求校验矩阵的秩（Z方向）
    @property
    def rank_z(self):
        return self.check_matrix_z.rank

    ##  USER：求校验矩阵（X方向）
    @property
    def check_matrix_x(self):
        matrix=Operator.get_matrix(self.generators_x,self.physical_number)
        matrix=matrix[:,0::2]
        return matrix

    ##  USER：求校验矩阵（Z方向）
    @property
    def check_matrix_z(self):
        matrix=Operator.get_matrix(self.generators_z,self.physical_number)
        matrix = matrix[:, 1::2]
        return matrix

    ##  USER：求码距（X方向）
    @property
    @abstractmethod
    def distance_x(self):
        pass

    ##  USER：求码距（Z方向）
    @property
    @abstractmethod
    def distance_z(self):
        return 1



import copy
import galois
import networkx as nx
import numpy as np
from extendedstim.Code.QuantumCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.QuantumCode.MajoranaCode import MajoranaCode
from extendedstim.Physics import MajoranaOperator


# %%  USER：实现fermionic lattice surgery
def MajoranaLatticeSurgery(code_A, code_B, index_A, index_B):
    code_A = code_A.copy()
    code_B = code_B.copy()
    assert isinstance(code_A, MajoranaCSSCode)
    assert isinstance(code_B, MajoranaCSSCode)
    assert isinstance(index_A, int)
    assert isinstance(index_B, int)

    # %%  SECTION：数据预处理
    majorana_number_A = code_A.physical_number
    majorana_number_B = code_B.physical_number
    code_B.index_map(np.arange(majorana_number_A, majorana_number_B+majorana_number_A))
    code_A.index_map(np.arange(majorana_number_A))
    logical_operator_0 = code_A.logical_operators_x[index_A]
    logical_operator_1 = code_B.logical_operators_x[index_B]
    support_index_vector_A = np.array(logical_operator_0.x_vector, dtype=int)
    support_index_vector_B = np.array(logical_operator_1.x_vector, dtype=int)

    ##  目标初始化
    check_origin_list = []
    check_ancilla_list = []
    check_modify_list = []
    check_stable_list = []
    check_gauge_list = []
    check_z_list = []
    check_measure_list = []

    vertex_qubit_list = []
    vertex_check_list = []

    edge_list = []
    for i in range(len(code_A.generators)):
        if len(code_A.generators[i].x_vector) > 0:
            if len(set(code_A.generators[i].x_vector) & set(support_index_vector_A)) > 0:
                check_origin_list.append(code_A.generators[i].copy())
                check_modify_list.append(code_A.generators[i].copy())
            else:
                check_stable_list.append(code_A.generators[i].copy())
        else:
            check_z_list.append(code_A.generators[i].copy())

    for i in range(len(code_B.generators)):
        if len(code_B.generators[i].x_vector) > 0:
            if len(set(code_B.generators[i].x_vector) & set(support_index_vector_B)) > 0:
                check_origin_list.append(code_B.generators[i].copy())
                check_modify_list.append(code_B.generators[i].copy())
            else:
                check_stable_list.append(code_B.generators[i].copy())
        else:
            check_z_list.append(code_B.generators[i].copy())

    ##  为待修改的stabilizers增加ancilla及其索引
    code = MajoranaCode()
    code.define_qubit(majorana_number_A + majorana_number_B)
    ancilla_list_list = []
    for i in range(len(check_modify_list)):
        x_vector_temp = check_modify_list[i].x_vector.tolist()
        z_vector_temp = []
        number_ancilla_temp = len(
            set(check_modify_list[i].x_vector) & set(np.append(support_index_vector_A, support_index_vector_B))) // 2
        temp = []
        for j in range(number_ancilla_temp):
            if check_modify_list[i].x_vector[0] < majorana_number_A:
                vertex_check_list.append((str(len(vertex_check_list)) + 'A'))
            else:
                vertex_check_list.append((str(len(vertex_check_list)) + 'B'))
            code.push_qubit(1)
            temp.append((code.qubit_list[-1], 'x'))
            temp.append((code.qubit_list[-1], 'z'))
            x_vector_temp.append(code.qubit_list[-1])
            z_vector_temp.append(code.qubit_list[-1])
            edge_list.append((str(code.qubit_list[-1]) + 'x', vertex_check_list[-1]))
            edge_list.append((str(code.qubit_list[-1]) + 'z', vertex_check_list[-1]))
        check_modify_list[i] = MajoranaOperator(x_vector_temp, z_vector_temp, 1)
        ancilla_list_list.append(temp)
    for i in range(majorana_number_A + majorana_number_B, code.number_qubit):
        vertex_qubit_list.append(str(i) + 'x')
        vertex_qubit_list.append(str(i) + 'z')

    # %%  SECTION：加入测量稳定子
    single_point = None
    single_qubit_list = []
    ##  右边logical更长的情况
    if len(support_index_vector_A) >= len(support_index_vector_B):
        ##  先将两边对齐的部分连起来
        for i in range(len(support_index_vector_B)):
            vertex_check_list.append((str(len(vertex_check_list)) + 'M'))
            x_vector_temp = [support_index_vector_A[i], support_index_vector_B[i]]
            z_vector_temp = []
            for j in range(len(check_modify_list)):
                if support_index_vector_A[i] in check_modify_list[j].x_vector or support_index_vector_B[i] in \
                        check_modify_list[j].x_vector:
                    temp = ancilla_list_list[j][-1]
                    if temp[1] == 'x':
                        x_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'x', vertex_check_list[-1]))
                    elif temp[1] == 'z':
                        z_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'z', vertex_check_list[-1]))
                    ancilla_list_list[j].pop(-1)

            if np.mod(len(x_vector_temp) + len(z_vector_temp), 2) != 0:
                if single_point is None:
                    code.push_qubit(1)
                    single_point = code.qubit_list[-1]
                    vertex_qubit_list.append(str(single_point) + 'x')
                    vertex_qubit_list.append(str(single_point) + 'z')
                    edge_list.append((str(single_point) + 'x', vertex_check_list[-1]))
                    single_qubit_list.append(code.qubit_list[-1])
                    x_vector_temp.append(code.qubit_list[-1])
                else:
                    z_vector_temp.append(single_point)
                    edge_list.append((str(single_point) + 'z', vertex_check_list[-1]))
                    single_point = None

            ##  引入新的measurement stabilizer
            check_measure_list.append(MajoranaOperator(x_vector_temp, z_vector_temp, 1))

        ##  将右边剩余的部分连起来
        length_B = len(support_index_vector_B)
        length_A = len(support_index_vector_A)
        for i in range((length_A - length_B) // 2):
            index_0 = length_B + 2 * i
            index_1 = index_0 + 1
            vertex_check_list.append((str(len(vertex_check_list)) + 'M'))
            x_vector_temp = [support_index_vector_A[index_0], support_index_vector_A[index_1]]
            z_vector_temp = []
            for j in range(len(check_modify_list)):
                if support_index_vector_A[index_0] in check_modify_list[j].x_vector:
                    temp = ancilla_list_list[j][-1]
                    if temp[1] == 'x':
                        x_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'x', vertex_check_list[-1]))
                    elif temp[1] == 'z':
                        z_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'z', vertex_check_list[-1]))
                    ancilla_list_list[j].pop(-1)
                if support_index_vector_A[index_1] in check_modify_list[j].x_vector:
                    temp = ancilla_list_list[j][-1]
                    if temp[1] == 'x':
                        x_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'x', vertex_check_list[-1]))
                    elif temp[1] == 'z':
                        z_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'z', vertex_check_list[-1]))
                    ancilla_list_list[j].pop(-1)

            if np.mod(len(x_vector_temp) + len(z_vector_temp), 2) != 0:
                if single_point is None:
                    code.push_qubit(1)
                    single_point = code.qubit_list[-1]
                    vertex_qubit_list.append(str(single_point) + 'x')
                    vertex_qubit_list.append(str(single_point) + 'z')
                    edge_list.append((str(single_point) + 'x', vertex_check_list[-1]))
                    single_qubit_list.append(code.qubit_list[-1])
                    x_vector_temp.append(code.qubit_list[-1])
                else:
                    z_vector_temp.append(single_point)
                    edge_list.append((str(single_point) + 'z', vertex_check_list[-1]))
                    single_point = None

            ##  引入新的measurement stabilizer
            check_measure_list.append(MajoranaOperator(x_vector_temp, z_vector_temp, 1))
    else:
        ##  先将两边对齐的部分连起来
        for i in range(len(support_index_vector_A)):
            vertex_check_list.append((str(len(vertex_check_list)) + 'M'))
            x_vector_temp = [support_index_vector_A[i], support_index_vector_B[i]]
            z_vector_temp = []
            for j in range(len(check_modify_list)):
                if support_index_vector_A[i] in check_modify_list[j].x_vector or support_index_vector_B[i] in \
                        check_modify_list[j].x_vector:
                    temp = ancilla_list_list[j][-1]
                    if temp[1] == 'x':
                        x_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'x', vertex_check_list[-1]))
                    elif temp[1] == 'z':
                        z_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'z', vertex_check_list[-1]))
                    ancilla_list_list[j].pop(-1)

            if np.mod(len(x_vector_temp) + len(z_vector_temp), 2) != 0:
                if single_point is None:
                    code.push_qubit(1)
                    single_point = code.qubit_list[-1]
                    vertex_qubit_list.append(str(single_point) + 'x')
                    vertex_qubit_list.append(str(single_point) + 'z')
                    edge_list.append((str(single_point) + 'x', vertex_check_list[-1]))
                    single_qubit_list.append(code.qubit_list[-1])
                    x_vector_temp.append(code.qubit_list[-1])
                else:
                    z_vector_temp.append(single_point)
                    edge_list.append((str(single_point) + 'z', vertex_check_list[-1]))
                    single_point = None

            ##  引入新的measurement stabilizer
            check_measure_list.append(MajoranaOperator(x_vector_temp, z_vector_temp, 1))

        ##  将右边剩余的部分连起来
        length_B = len(support_index_vector_B)
        length_A = len(support_index_vector_A)
        for i in range((length_B - length_A) // 2):
            index_0 = length_A + 2 * i
            index_1 = index_0 + 1
            vertex_check_list.append((str(len(vertex_check_list)) + 'M'))
            x_vector_temp = [support_index_vector_B[index_0], support_index_vector_B[index_1]]
            z_vector_temp = []
            for j in range(len(check_modify_list)):
                if support_index_vector_B[index_0] in check_modify_list[j].x_vector or support_index_vector_B[
                    index_1] in check_modify_list[j].x_vector:
                    temp = ancilla_list_list[j][-1]
                    if temp[1] == 'x':
                        x_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'x', vertex_check_list[-1]))
                    elif temp[1] == 'z':
                        z_vector_temp.append(temp[0])
                        edge_list.append((str(temp[0]) + 'z', vertex_check_list[-1]))
                    ancilla_list_list[j].pop(-1)

            if np.mod(len(x_vector_temp) + len(z_vector_temp), 2) != 0:
                if single_point is None:
                    code.push_qubit(1)
                    single_point = code.qubit_list[-1]
                    vertex_qubit_list.append(str(single_point) + 'x')
                    vertex_qubit_list.append(str(single_point) + 'z')
                    edge_list.append((str(single_point) + 'x', vertex_check_list[-1]))
                    single_qubit_list.append(code.qubit_list[-1])
                    x_vector_temp.append(code.qubit_list[-1])
                else:
                    z_vector_temp.append(single_point)
                    edge_list.append((str(single_point) + 'z', vertex_check_list[-1]))
                    single_point = None

            ##  引入新的measurement stabilizer
            check_measure_list.append(MajoranaOperator(x_vector_temp, z_vector_temp, 1))

    # %%  SECTION：图论计算规范稳定子
    for i in range(code.number_qubit - majorana_number_A - majorana_number_B):
        temp = MajoranaOperator([i + majorana_number_B + majorana_number_A], [i + majorana_number_B + majorana_number_A], 1)
        check_ancilla_list.append(temp)

    ##  获取关键参数

    for i in range(len(single_qubit_list)):
        vertex_check_list.append(str(i) + 'D')
        edge_list.append((str(single_qubit_list[i]) + 'x', str(i) + 'D'))
        edge_list.append((str(single_qubit_list[i]) + 'z', str(i) + 'D'))

    graph = nx.Graph()
    graph.add_nodes_from(vertex_check_list)
    graph.add_nodes_from(vertex_qubit_list)
    graph.add_edges_from(edge_list)
    check_gauge_list = IndependentCyclesFinder(graph)

    target = logical_operator_0.mul(logical_operator_1, code.number_qubit)

    # %%  SECTION：返回结果

    code.check_list = check_stable_list + check_modify_list+ check_gauge_list + check_z_list+check_measure_list
    code.number_checker = len(code.check_list)
    code.target=target
    assert code.commute_judge()  # 检查对易性
    return code


def ZechuanLatticeSurgery(code_ldpc, index)->MajoranaCode:

    #%%  SECTION：数据标准化
    code_ldpc=code_ldpc.copy()
    assert isinstance(code_ldpc, MajoranaCSSCode)
    assert isinstance(index, int)


    #%%  SECTION：数据预处理
    number_qubit_ldpc=code_ldpc.number_qubit
    logical_operator_ldpc=code_ldpc.logical_operator_list_x[index]
    support_index_vector_ldpc = np.array(logical_operator_ldpc.x_vector, dtype=int)
    code_color=MajoranaCode.ColorCode(len(support_index_vector_ldpc))
    number_qubit_color=code_color.number_qubit
    logical_operator_color = code_color.logical_operator_list_x[0]
    code_color.index_map(number_qubit_ldpc+number_qubit_color,np.arange(number_qubit_color)+number_qubit_ldpc)
    code_ldpc.index_map(number_qubit_ldpc+number_qubit_color,np.arange(number_qubit_ldpc))
    support_index_vector_color=np.array(logical_operator_color.x_vector,dtype=int)
    support_index_vector_color=support_index_vector_color.tolist()
    support_index_vector_ldpc=support_index_vector_ldpc.tolist()

    ##  提取与这些费米子相关联的校验子
    check_list_x_fix_ldpc=[]
    check_list_x_unfix_ldpc=[]
    check_list_z_ldpc=[]
    check_list_x_fixed_ldpc=[]
    check_list_x_fix_color=[]
    check_list_x_unfix_color=[]
    check_list_z_color=[]
    check_list_x_fixed_color=[]

    for i in range(len(code_ldpc.check_list)):
        if len(code_ldpc.check_list[i].x_vector)>0:
            if len(set(code_ldpc.check_list[i].x_vector)&set(support_index_vector_ldpc))>0:
                check_list_x_fix_ldpc.append(code_ldpc.check_list[i].copy())
            else:
                check_list_x_unfix_ldpc.append(code_ldpc.check_list[i].copy())
        else:
            check_list_z_ldpc.append(code_ldpc.check_list[i].copy())

    for i in range(len(code_color.check_list)):
        if len(code_color.check_list[i].x_vector)>0:
            if len(set(code_color.check_list[i].x_vector)&set(support_index_vector_color))>0:
                check_list_x_fix_color.append(code_color.check_list[i].copy())
            else:
                check_list_x_unfix_color.append(code_color.check_list[i].copy())
        else:
            check_list_z_color.append(code_color.check_list[i].copy())

    code=MajoranaCode()
    code.define_qubit(number_qubit_ldpc+number_qubit_color)
    code.check_origin_list=check_list_x_fix_color+check_list_x_fix_ldpc
    ##  记录与support关联的ancilla的索引
    ancilla_dict_ldpc={}
    ancilla_dict_color={}
    for i in range(len(support_index_vector_ldpc)):
        ancilla_dict_ldpc[support_index_vector_ldpc[i]]=[]
        ancilla_dict_color[support_index_vector_color[i]]=[]

    ##  对color code索引
    check_index_color=code_color.logical_plaqutte=[1,0,2,3]
    single_plaqutte=[0,1,2,3]
    def find_plaqutte(index_0,index_1):
        op_0=check_list_x_fix_color[0]
        op_1=check_list_x_fix_color[1]
        op_2=check_list_x_fix_color[2]
        op_3=check_list_x_fix_color[3]
        if index_0==0 and index_1==1:
            return op_1,[1],None
        elif index_0==0 and index_1==2:
            return op_1.mul(op_0,code.number_qubit),[0,1],None
        elif index_0==0 and index_1==3:
            return op_1.mul(op_2,code.number_qubit),[1,2],[support_index_vector_color[1],support_index_vector_color[2]]
        elif index_0==0 and index_1==4:
            return op_1.mul(op_3,code.number_qubit),[1,3],[support_index_vector_color[1],support_index_vector_color[3]]
        elif index_0==1 and index_1==2:
            return op_0,[0],None
        elif index_0==1 and index_1==3:
            return op_0.mul(op_2,code.number_qubit),[0,2],None
        elif index_0==1 and index_1==4:
            return op_0.mul(op_3,code.number_qubit),[0,3],[support_index_vector_color[2],support_index_vector_color[3]]
        elif index_0==2 and index_1==3:
            return op_2,[2],None
        elif index_0==2 and index_1==4:
            return op_2.mul(op_3,code.number_qubit),[2,3],None
        elif index_0==3 and index_1==4:
            return op_3,[3],None

    check_list_gauge=[]
    ##  对support组队
    couple_list=[]
    for i in range(len(check_list_x_fix_ldpc)):
        temp=set(check_list_x_fix_ldpc[i].x_vector) & set(support_index_vector_ldpc)
        temp=list(temp)
        couple_list.append([])
        for j in range(len(temp)//2):
            couple_list[i].append((temp[2*j],temp[2*j+1]))

    ##  修改LDPC code的check
    for i in range(len(check_list_x_fix_ldpc)):
        for j in range(len(couple_list[i])):
            x_vector_temp=check_list_x_fix_ldpc[i].x_vector.copy()
            z_vector_temp=check_list_x_fix_ldpc[i].z_vector.copy()
            code.push_qubit(1)
            ancilla_dict_ldpc[couple_list[i][j][0]].append((code.qubit_list[-1],'x'))
            ancilla_dict_ldpc[couple_list[i][j][1]].append((code.qubit_list[-1],'z'))
            x_vector_temp=np.append(x_vector_temp,code.qubit_list[-1])
            z_vector_temp=np.append(z_vector_temp,code.qubit_list[-1])
            check_list_x_fix_ldpc[i]=MajoranaOperator(x_vector_temp, z_vector_temp,1)
            index_0=couple_list[i][j][0]
            index_1=couple_list[i][j][1]
            index_min=support_index_vector_ldpc.index(index_0)
            index_max=support_index_vector_ldpc.index(index_1)
            if index_min>index_max:
                temp=index_min
                index_min=index_max
                index_max=temp
            operator_color_temp,index_list,associate_list=find_plaqutte(index_min,index_max)
            x_vector_temp = operator_color_temp.x_vector.copy()
            z_vector_temp = operator_color_temp.z_vector.copy()
            for i in range(len(index_list)):
                single_plaqutte[index_list[i]]=None
            if associate_list is not None:
                index_temp_0=associate_list[0]
                index_temp_1=associate_list[1]
                code.push_qubit(2)

                ancilla_dict_color[index_temp_0].append((code.qubit_list[-1],'x'))
                ancilla_dict_color[index_temp_0].append((code.qubit_list[-2],'x'))
                ancilla_dict_color[index_temp_1].append((code.qubit_list[-1],'z'))
                ancilla_dict_color[index_temp_1].append((code.qubit_list[-2],'z'))
                x_vector_temp_gauge=[code.qubit_list[-2],code.qubit_list[-1]]
                z_vector_temp_gauge=[code.qubit_list[-2],code.qubit_list[-1]]
                check_list_gauge.append(MajoranaOperator(x_vector_temp_gauge, z_vector_temp_gauge, 1))
                x_vector_temp=np.append(x_vector_temp,code.qubit_list[-2])
                z_vector_temp=np.append(z_vector_temp,code.qubit_list[-2])
            index_0_color = support_index_vector_color[support_index_vector_ldpc.index(index_0)]
            index_1_color = support_index_vector_color[support_index_vector_ldpc.index(index_1)]
            code.push_qubit(1)
            x_vector_temp=np.append(x_vector_temp,code.qubit_list[-1])
            z_vector_temp=np.append(z_vector_temp,code.qubit_list[-1])
            ancilla_dict_color[index_0_color].append((code.qubit_list[-1],'x'))
            ancilla_dict_color[index_1_color].append((code.qubit_list[-1],'z'))
            check_list_x_fixed_color.append(MajoranaOperator(x_vector_temp, z_vector_temp,1))
            if associate_list is not None:
                x_vector_temp=[code.qubit_list[-4],code.qubit_list[-1]]
                z_vector_temp=[code.qubit_list[-4],code.qubit_list[-1]]
            else:
                x_vector_temp=[code.qubit_list[-2],code.qubit_list[-1]]
                z_vector_temp=[code.qubit_list[-2],code.qubit_list[-1]]
            check_list_gauge.append(MajoranaOperator(x_vector_temp, z_vector_temp,1))

    for i in range(len(single_plaqutte)):
        if single_plaqutte[i] is not None:
            operator_color_temp=check_list_x_fix_color[single_plaqutte[i]]
            overlap=list(set(operator_color_temp.x_vector)&set(support_index_vector_color))
            x_vector_temp = operator_color_temp.x_vector.copy()
            z_vector_temp = operator_color_temp.z_vector.copy()
            code.push_qubit(2)
            ancilla_dict_color[overlap[0]].append((code.qubit_list[-1], 'x'))
            ancilla_dict_color[overlap[0]].append((code.qubit_list[-2], 'x'))
            ancilla_dict_color[overlap[1]].append((code.qubit_list[-1], 'z'))
            ancilla_dict_color[overlap[1]].append((code.qubit_list[-2], 'z'))
            x_vector_temp=np.append(x_vector_temp,code.qubit_list[-2])
            z_vector_temp=np.append(z_vector_temp,code.qubit_list[-2])
            check_list_x_fixed_color.append(MajoranaOperator(x_vector_temp, z_vector_temp, 1))
            x_vector_temp = [code.qubit_list[-2], code.qubit_list[-1]]
            z_vector_temp = [code.qubit_list[-2], code.qubit_list[-1]]
            check_list_gauge.append(MajoranaOperator(x_vector_temp, z_vector_temp, 1))
    check_list_measurement = []
    for i in range(len(support_index_vector_ldpc)):
        x_vector_temp=[support_index_vector_ldpc[i],support_index_vector_color[i]]
        z_vector_temp=[]
        for j in range(len(ancilla_dict_ldpc[support_index_vector_ldpc[i]])):
            if ancilla_dict_ldpc[support_index_vector_ldpc[i]][j][1]=='x':
                x_vector_temp=np.append(x_vector_temp,ancilla_dict_ldpc[support_index_vector_ldpc[i]][j][0])
            else:
                z_vector_temp = np.append(z_vector_temp, ancilla_dict_ldpc[support_index_vector_ldpc[i]][j][0])
        for j in range(len(ancilla_dict_color[support_index_vector_color[i]])):
            if ancilla_dict_color[support_index_vector_color[i]][j][1]=='x':
                x_vector_temp=np.append(x_vector_temp,ancilla_dict_color[support_index_vector_color[i]][j][0])
            else:
                z_vector_temp = np.append(z_vector_temp, ancilla_dict_color[support_index_vector_color[i]][j][0])
        check_list_measurement.append(MajoranaOperator(x_vector_temp, z_vector_temp, 1))
    check_list_x_fixed_ldpc=check_list_x_fix_ldpc
    code.check_list=check_list_x_unfix_ldpc+check_list_x_unfix_color
    code.check_stable_list=check_list_x_unfix_ldpc+check_list_x_unfix_color+check_list_z_ldpc+check_list_z_color
    code.check_list+=check_list_z_ldpc+check_list_z_color
    code.check_list+=check_list_x_fixed_color+check_list_x_fixed_ldpc
    code.check_fixed_list=check_list_x_fixed_color+check_list_x_fixed_ldpc
    code.check_gauge_list=check_list_gauge
    code.check_list+=check_list_gauge
    code.ancilla_list=[]
    code.stop=len(code.check_list)
    code.check_list+=check_list_measurement
    code.check_measure_list=check_list_measurement
    for i in range(number_qubit_ldpc+number_qubit_color,code.number_qubit):
        temp=MajoranaOperator([i],[i],1)
        code.ancilla_list.append(temp)
    code.number_checker=len(code.check_list)

    #%%  SECTION：返回结果

    assert code.commute_judge()  # 检查对易性
    return code


def IndependentCyclesFinder(G):
    qubit_list=[]
    check_list=[]
    for index,value in enumerate(G.nodes()):
        if value[-1]=='A' or value[-1]=='B' or value[-1]=='D' or value[-1]=='M':
            check_list.append(value)
        else:
            qubit_list.append(value)
    matrix=np.zeros((len(check_list),len(qubit_list)),dtype=int)
    for index0,check in enumerate(check_list):
        for index1,qubit in enumerate(qubit_list):
            if (qubit,check) in G.edges():
                matrix[index0,index1]=1
    GF=galois.GF(2**1)
    matrix=GF(matrix)
    result=matrix.null_space()
    check_gauge_list=[]
    for i in range(len(result)):
        index_list=np.where(result[i]!=0)[0]
        temp_x=[]
        temp_z=[]
        for j in range(len(index_list)):
            temp=qubit_list[index_list[j]]
            if temp[-1]=='x':
                temp_x.append(int(temp[:-1]))
            elif temp[-1]=='z':
                temp_z.append(int(temp[:-1]))
            else:
                raise ValueError
        check_gauge_list.append(MajoranaOperator(temp_x,temp_z,1))
    return check_gauge_list


import galois
import numpy as np
from extendedstim.Code.LinearCode.LinearCode import LinearCode
from extendedstim.tools import isinteger


class BicycleCode(LinearCode):
    #%%  USER：构造方法
    def __init__(self, N, k, M, seed):

        ##  ———数据预处理---
        assert isinteger(N), "N必须是整数"
        assert isinteger(k), "k必须是整数"
        assert isinteger(M), "M必须是整数"
        assert isinteger(seed), "seed必须是整数"

        ##  ---生成校验矩阵---
        np.random.seed(seed)
        H, diff_set = generate_matrix(N, k, M)
        assert np.all(H @ H.T == 0)
        super().__init__(H)


#%%  KEY：生成循环矩阵
def create_circulant_matrix(difference_set, size):
    matrix = np.zeros((size, size), dtype=int)
    for i in range(size):
        for d in difference_set:
            j = (i + d) % size
            matrix[i, j] = 1
    return matrix


#%%  KEY：生成差集
def generate_difference_set(n, k):
    """生成满足唯一差值特性的差集"""
    # 初始化差集
    diff_set = [0]
    # 记录已使用的差值
    used_differences = set()

    while len(diff_set) < k:
        candidate = np.random.randint(1, n - 1)
        valid = True

        # 检查与现有元素的所有差值是否唯一
        for d in diff_set:
            diff1 = (candidate - d) % n
            diff2 = (d - candidate) % n

            if diff1 in used_differences or diff2 in used_differences:
                valid = False
                break

        if valid:
            # 添加新元素并记录差值
            diff_set.append(candidate)
            for d in diff_set[:-1]:
                diff1 = (candidate - d) % n
                diff2 = (d - candidate) % n
                used_differences.add(diff1)
                used_differences.add(diff2)

    return sorted(diff_set)


#%%  KEY：生成循环矩阵
def generate_matrix(N, k, M):
    # 验证参数
    if N % 2 != 0:
        raise ValueError("N必须是偶数")
    if k % 2 != 0:
        raise ValueError("k必须是偶数")
    if M >= N / 2:
        raise ValueError("M必须小于N/2")

    n = N // 2  # 循环矩阵大小
    k_half = k // 2  # C的行权重

    # 步骤1: 生成满足唯一差值特性的差集
    difference_set = generate_difference_set(n, k_half)

    # 步骤2: 创建循环矩阵C和其转置
    C = create_circulant_matrix(difference_set, n)
    C_T = C.T

    # 步骤3: 构造H0 = [C, C_T]
    H0 = np.hstack((C, C_T))

    # 步骤4: 删除行以实现均匀列权重
    rows_to_keep = list(range(H0.shape[0]))

    # 计算初始列权重
    col_weights = np.sum(H0, axis=0)

    # 计算需要删除的行数
    rows_to_delete = H0.shape[0] - M

    # 贪心算法删除行，使列权重均匀
    for _ in range(rows_to_delete):
        best_row = -1
        best_variance = float('inf')

        # 尝试删除每一行，找到使列权重方差最小的行
        for row in rows_to_keep:
            # 临时删除该行
            temp_weights = col_weights - H0[row, :]
            variance = np.var(temp_weights)

            if variance < best_variance:
                best_variance = variance
                best_row = row

        # 删除最佳行
        rows_to_keep.remove(best_row)
        col_weights -= H0[best_row, :]

    # 创建最终矩阵
    H = H0[rows_to_keep, :]

    # 转换为GF(2)矩阵
    GF2 = galois.GF2
    H_gf2 = GF2(H)

    return H_gf2, difference_set
